%---------------------- Median polish -------------------------------------
%  Median Polish. Median is more resistant to outliers in the data than mean, it
%  is robust to outliers. Each value in an additive model is decomposed to
%  several parts:
% 
%       data = overllEffects + rowEffects + columnEffects + residual
% 
%  The median polish algorithm consisted these steps: 
%   1. Subtract the medians of each row from the row values, record the row
%      medians in the row effect. 
%   2. Subtract the medians of each column from the column values, and record
%      the column medians in the column effect. 
%   3. Repeat these steps until it converges, i.e. the medians of both rows and
%      columns are 0.0 (+/- 0.5). The function allow for up to 10 iterations. 
%   4. Compute the median of row median and add it to the overall effect,
%      subtract it from the row median. Then do the same for the column effect.
% 
%  The function returns the overall and column effects as the expression measure
%  for a given probe set.

function avgRet = doMedianPolish(ppvalues)
[rows, cols] = size(ppvalues);

% log2 transform
%z = log2(ppvalues);
z = ppvalues;

% % medFunc = @median;
% % if any(isnan(ppvalues))
% %     medFunc = @nanmedian;
% % end

% Define maximum number of iteration if the median polish procedure does not
% converge
maxIteration = 10; % This is the number RMAExpress uses
tol = 0.01;
allEffect = 0.0;
rowEffects = zeros(rows,1); % For row (chip) effects 
colEffects = zeros(1,cols); % For col (probe) effects

iloop = 1;
row_median = median(z, 2);
while iloop <= maxIteration
    % subtract the matrix by the row median
    tmp = repmat(row_median, 1, cols);
    z = z - tmp;

    % Add to row effect
    rowEffects = rowEffects + row_median;
    % Get column median
    col_median = median(z, 1);
    
    % Subtract the matrix by the column median
    tmp = repmat(col_median, rows, 1);
    z = z - tmp;

    %Add to column effect
    colEffects = colEffects + col_median;
    % Get row median
    row_median = median(z, 2);
    
    % check for convergence
    if all(abs(row_median) < tol ) && all(abs(median(z, 1)) < tol )
        break;
    end
     
    iloop = iloop+1;
end

% The commented out code below is the formal way of summing up the effects,
% since the row effect will not be added to the final results, so we can cut a
% few steps.
% Add median of the row effect to the all effect
me = median(rowEffects);

allEffect = allEffect + me;
% % Subtract it from row effect
 rowEffects = rowEffects - me
% 
% % Add median of the col effect to the all effect
 me = median(colEffects);
 allEffect = allEffect + me;
 % Subtract it from col effect
 colEffects = colEffects - me

% The total probe (row effect) should be zero
avgRet = colEffects + allEffect;
