/* absolute_rank_selection.c: Implementations of functions for
 * computing normalising functions based on absolute rank selection
 * invariant sets. The two main functions compute normalising
 * functions for all possible thresholds and for a specific threshold,
 * respectively.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "absolute_rank_selection.h"
#include "mergesort.h"
#include "wrappers.h"

/* xmalloc(n): wrapper function for malloc, testing whether memory was
 * correctly allocated.
 */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "absolute_rank_selection");
#else
  return wmalloc(n);
#endif
}

/* cmp_idpair(a, b): return true if value of a is less than value of b */
static int cmp_idpair(_ARS_idpair *a, _ARS_idpair *b)
{
  return (a->value < b->value);
}

/* cmp_iipair(a, b): return true if value of a is less than value of b */
static int cmp_iipair(_ARS_iipair *a, _ARS_iipair *b)
{
  return (a->value < b->value);
}

/* rank_difference(n, x, y): return an array with rank differences for
 * all points when sorted according to x and y respectively.
 */
static int *rank_difference(int n, double *x, double *y)
{
  _ARS_idpair *sort = (_ARS_idpair *)xmalloc(n * sizeof(_ARS_idpair));
  int *rd = (int *)xmalloc(n * sizeof(int));
  int i;

  /* Sort elements according to x value */
  for (i = 0; i < n; i++){
    sort[i].index = i;
    sort[i].value = x[i];
  }
  mergesort((void *)sort, n, sizeof(_ARS_idpair),
	    (int (*)(void *, void *))cmp_idpair);

  /* Store rank from x value sorting and prepare for sort on y value */
  for (i = 0; i < n; i++){
    rd[sort[i].index] = i;
    sort[i].value = y[sort[i].index];
  }

  /* Sort elements according to y value */
  mergesort((void *)sort, n, sizeof(_ARS_idpair),
	    (int (*)(void *, void *))cmp_idpair);

  /* Compute rank difference */
  for (i = 0; i < n; i++)
    rd[sort[i].index] = abs(rd[sort[i].index] - i);

  /* Clean up */
  wfree(sort);

  return rd;
}

/* absolute_rank_selection_all(n, x, y, new_function): find
 * normalising functions returned by new_function when only points
 * with an absolute rank difference of a given threshold are used for
 * all possible thresholds. Only for thresholds where more points are
 * included will the corresponding entry in the array of normalising
 * functions returned contain an actual function, all other entries in
 * the array are NULL. If new_function is NULL the invariant sets are
 * returned instead of functions inferred from the invariant set.
 */
void **absolute_rank_selection_all(int n, double *x, double *y,
				   _ARS_newfunction new_function)
{
  _ARS_iipair *sort = (_ARS_iipair *)xmalloc(n * sizeof(_ARS_iipair));
  int *rd = rank_difference(n, x, y);
  double *xs, *ys;
  int i;
  void **fs = (void **)xmalloc(n * sizeof(void *));

  /* Sort the rank differences of the points, and initialise array of
   * normalising functions.
   */
  for (i = 0; i < n; i++){
    sort[i].index = i;
    sort[i].value = rd[i];
    fs[i] = NULL;
  }
  wfree(rd);
  mergesort((void *)sort, n, sizeof(_ARS_iipair),
	    (int (*)(void *, void *))cmp_iipair);

  if (new_function != NULL){
    /* Store the points in xs, ys in order of increasing rank
     * difference, and compute the next normalising function whenever
     * the rank difference increases.
     */
    xs = (double *)xmalloc(n * sizeof(double));
    ys = (double *)xmalloc(n * sizeof(double));
    xs[0] = x[sort[0].index];
    ys[0] = y[sort[0].index];
    for (i = 1; i < n; i++){
      if (sort[i].value > sort[i - 1].value)
	/* Change in rank difference - compute normalising function for
	 * all preceeding points.
	 */
	fs[sort[i - 1].value] = new_function(i, xs, ys);
      xs[i] = x[sort[i].index];
      ys[i] = y[sort[i].index];
    }
    /* Compute normalising function when all points are included */
    fs[sort[n - 1].value] = new_function(n, xs, ys);

    /* Clean up */
    wfree(xs);
    wfree(ys);
  }
  else{
    /* Store arrays in fs where the t'th array has i'th element equal
     * to 1 if the rank difference for the i'th point is at most t, 0
     * otherwise; if the t'th and t - 1'st array would be identical,
     * the t'th element of fs is NULL.
     */
    /* Start with initialising all entries to 0 */
    xs = (double *)xmalloc(n * sizeof(double));
    for (i = 0; i < n; i++)
      xs[i] = 0;
    /* Change values to 1 in order of increasing rank difference;
     * whenever the rank difference increases, store the array at the
     * right position in fs and allocate a new array with the same
     * content.
     */
    xs[sort[0].index] = 1;
    for (i = 1; i < n; i++){
      if (sort[i].value != sort[i - 1].value){
	/* Change in rank difference - store current array in fs */
	ys = (double *)xmalloc(n * sizeof(double));
	memcpy(ys, xs, n * sizeof(double));
	fs[sort[i - 1].value] = (void *)ys;
      }
      xs[sort[i].index] = 1;
    }
    /* We haven't stored last invariant set (i.e. the set consisting
     * of all points) in fs yes.
     */
    fs[sort[i].value] = (void *)xs;
  }

  /* Clean up */
  wfree(sort);

  return fs;
}

/* absolute_rank_selection(n, x, y, t, new_function): compute
 * normalising function by passing the points whose rank differs by at
 * most t when sorted according to x values and y values,
 * respectively, to new_function. If new_function is NULL the
 * invariant set are returned instead of the function inferred from
 * the invariant set.
 */
void *absolute_rank_selection(int n, double *x, double *y, int t,
			      _ARS_newfunction new_function)
{
  void *f;
  int *rd = rank_difference(n, x, y);
  double *xs, *ys;
  int i, j, m;

  if (new_function != NULL){
    /* Find number of points with absolute rank difference at most t */
    m = 0;
    for (i = 0; i < n; i++)
      m += (rd[i] <= t);

    /* Copy points with rank difference at most t to xs, ys */
    xs = (double *)xmalloc(m * sizeof(double));
    ys = (double *)xmalloc(m * sizeof(double));
    i = 0;
    for (j = 0; j < m; j++){
      /* Skip point with rank difference more than t */
      while (rd[i] > t) i++;
      xs[j] = x[i];
      ys[j] = y[i];
      i++;
    }

    /* Compute normalising function based on points xs, ys */
    f = new_function(m, xs, ys);

    /* Clean up */
    wfree(rd);
    wfree(xs);
    wfree(ys);
    return f;
  }
  else{
    /* Create array with i'th element 1 if the rank difference for the
     * i'th point is at most t, 0 otherwise.
     */
    xs = (double *)xmalloc(n * sizeof(double));
    for (i = 0; i < n; i++)
      xs[i] = (rd[i] <= t);

    return (void *)xs;
  }
}
