/* iploess.c: Implementation of functions to do one dimensional
 * interpolated loess fitting.
 */
/* Rune Lyngs, June 2003 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include "iploess.h"
#include "loess.h"
#include "polynomial.h"
#include "mergesort.h"
#include "rankselect.h"
#include "wrappers.h"

/* xmalloc(n): Wrapper function for malloc checking return value */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "iploess");
#else
  return wmalloc(n);
#endif
}

/* iploess(l): returns an interpolated approximation of the Loess
 * structure l, approximating the Loess fit by different polynomials
 * for each of n equal sized intervals between a and b, as well as a
 * polynomial for values smaller than a and a polynomial for values
 * larger than b. The polynomial for a specific interval is of degree
 * equal to the lambda value of l, and is determined as the unique
 * polynomial agreeing with l on lambda + 1 equally spaced points
 * starting at the left boundary of the interval and ending at the
 * right boundary of the interval. For the approximate fits outside
 * the range from a to b to work well, a has to be (significantly)
 * smaller than the smallest x value of any inference point in l, and
 * b has to be (significantly) larger than the largest x value of any
 * inference point in l.
 */
IPLoess *iploess(Loess *l, int n, double a, double b)
{
  int i, j;
  IPLoess *ipl = (IPLoess *)xmalloc(sizeof(IPLoess));
  double *x = (double *)xmalloc((l->lambda + 1) * sizeof(double)),
    *y = (double *)xmalloc((l->lambda + 1) * sizeof(double));
  Polynomial *p;

  /* Sanity check */
  if (b <= a){
    fprintf(stderr, "Attempt to find approximating Loess fit for interval of non-positive length.\n");
    return NULL;
  }

  /* Initialise new IPLoess structure */
  ipl->n = n;
  ipl->p = (double **)xmalloc((n + 2) * sizeof(double *));
  for (i = 0; i < n + 2; i++)
    ipl->p[i] = (double *)xmalloc((l->lambda + 1) * sizeof(double));
  ipl->left = a;
  ipl->right = b;
  ipl->lambda = l->lambda;

  /* Determine approximating polynomial for each interval */
  /* Leftmost point in a given interval corresponds to the rightmost
   * point in the interval to its left; except for the first interval,
   * so we start by computing the value at the leftmost point of the
   * first interval seperately.
   */
  x[ipl->lambda] = a;
  y[ipl->lambda] = loess_evaluate(l, a);
  for (i = 1; i <= n; i++){
    /* Compute value of Loess fit at lambda + 1 equidistant points
     * covering this interval.
     */
    for (j = 0; j < ipl->lambda; j++){
      x[j] = a + i * (b - a) / n - j * (b - a) / (n * ipl->lambda);
      y[j] = loess_evaluate(l, x[j]);
    }
    /* Find unique polynomial of degree lambda agreeing with these points */
    p = polynomial_fit(ipl->lambda + 1, x, y);
    /* Copy it to ipl->p */
    for (j = 0; j <= ipl->lambda; j++)
      ipl->p[i][j] = p->a[j];

    /* Clean up */
    polynomial_free(p);

    /* Rightmost point in this interval equals leftmost point of next
     * interval, so we copy the computed value at this point
     * accordingly.
     */
    x[ipl->lambda] = x[0];
    y[ipl->lambda] = y[0];
  }

  /* Determine approximating polynomial for values smaller than a; for
   * a point with an x value much smaller than any of the inference
   * points of the Loess structure, the Loess fit will be the
   * polynomial fit obtained by the alpha inference points with
   * smallest x value, ignoring the tricube weighting.
   */
  if (l->w != NULL)
    p = polynomial_wlsdpolynomial(ipl->lambda + 1, l->q, l->x, l->y, l->w, 0,
				  a, b);
  else
    p = polynomial_lsdpolynomial(ipl->lambda + 1, l->q, l->x, l->y, 0, a, b);
  if (p != NULL){
    for (i = 0; i <= ipl->lambda; i++)
      ipl->p[0][i] = p->a[i];
    polynomial_free(p);
  }
  else{
    fprintf(stderr, "Warning: Could not convert Loess fit to approximate Loess fit.\n");
    iploess_free(ipl);
    return NULL;
  }

  /* Now do the same thing for values larger than b; only difference
   * is that we use the alpha inference points with largest x
   * value.
   */
  if (l->w != NULL)
    p = polynomial_wlsdpolynomial(ipl->lambda + 1, l->q, l->x + l->n - l->q,
				  l->y + l->n - l->q, l->w + l->n - l->q, 0,
				  a, b);
  else
    p = polynomial_lsdpolynomial(ipl->lambda + 1, l->q, l->x + l->n - l->q,
				 l->y + l->n - l->q, 0, a, b);
  if (p != NULL){
    for (i = 0; i <= ipl->lambda; i++)
      ipl->p[n + 1][i] = p->a[i];
    polynomial_free(p);
  }
  else{
    fprintf(stderr, "Warning: Could not convert Loess fit to approximate Loess fit.\n");
    iploess_free(ipl);
    return NULL;
  }

  /* Clean up */
  wfree(x);
  wfree(y);

  return ipl;
}

/* square(a); returns a squared */
static double square(double a)
{
  return a * a;
}

/* bisquare(u, b): computes the bisquare function */
static double bisquare(double u, double b)
{
  if (u < b)
    return square(1 - square(u / b));
  else
    return 0;
}

/* compute_q(n, alpha, lambda): returns the number of points used for
 * function inference with n datapoints and parameters alpha and
 * lambda.
 */
static int compute_q(int n, double alpha, int lambda)
{
  int q = floor(n * alpha);

  /* The number should be larger than the degree of the polynomial,
   * but no larger than the total number of points.
   */
  if (q < lambda + 1)
    q = lambda + 1;
  if (q > n)
    q = n;

  return q;
}

/* cmp_doublep(x, y): compares the doubles pointed to by x and y */
static int cmp_doublep(double *x, double *y)
{
  return *x < *y;
}

/* iploess_robust(l, n, a, b): similar to iploess, except that the
 * Loess fit is robustified. As it might be too time consuming to just
 * robustify l by loess_robustify (as this requires evaluating l at
 * all inference points several times) and then compute the
 * approximate Loess fit by iploess, the robustifying reweighting has
 * to be carried out using intermediate approximate Loess fits.
 */
IPLoess *iploess_robust(Loess *l, int n, double a, double b)
{
  IPLoess *ipl;
  Loess *lo = (Loess *)xmalloc(sizeof(Loess));
  int i, j, k, c;
  double m, s, *epsilon = (double *)xmalloc(l->n * sizeof(double));


  /* Create a temporary Loess structure, initialise to copy of l */
  lo->x = (double *)xmalloc(l->n * sizeof(double));
  lo->y = (double *)xmalloc(l->n * sizeof(double));
  lo->w = (double *)xmalloc(l->n * sizeof(double));
  lo->alpha = l->alpha;
  lo->lambda = l->lambda;
  lo->q = l->q;
  lo->n = l->n;
  for (i = 0; i < l->n; i++){
    lo->x[i] = l->x[i];
    lo->y[i] = l->y[i];
    lo->w[i] = (l->w != NULL ? l->w[i] : 1);
  }

  /* Iteratively update the weights */
  j = 0;
  do{
    /* Compute an approximate Loess fit from temporary Loess fit */
    ipl = iploess(lo, n, l->x[0], l->x[l->n - 1]);
    if (ipl == NULL){
      fprintf(stderr, "Warning: Could not compute robust approximate Loess fit from Loess fit.\n");
      loess_free(lo);
      return NULL;
    }

    /* Compute deviation to current estimate for each point in l */
    k = 0;
    for (i = 0; i < l->n; i++){
      epsilon[i] = iploess_evaluate(ipl, l->x[i]);
      if (epsilon[i] != DBL_MAX)
	epsilon[i] = fabs(epsilon[i] - l->y[i]);
      else
	/* Could not evaluate the loess curve at i'th point */
	k++;
    }
    iploess_free(ipl);
    /* Compute robustness weights; we rescale the robustness weights
     * to have average 1, so we also need to compute the sum of the
     * robustness weights.
     */
    m = 6 * epsilon[rankselect(epsilon, (l->n - k) / 2, l->n, sizeof(double),
			       (int (*)(void *, void *))cmp_doublep)];
    s = 0;
    for (i = 0; i < l->n; i++){
      if (epsilon[i] != DBL_MAX){
	epsilon[i] = bisquare(epsilon[i], m);
	s += epsilon[i];
      }
      else
	epsilon[i] = 0;
    }

    if (s == 0){
      /* All weights are 0 - abort */
      loess_free(lo);
      wfree(epsilon);
      return NULL;
    }

    /* Update weights and check for convergence */
    c = 0;
    k = 0;
    for (i = 0; i < l->n; i++){
      /* Compute new weight for point i */
      m = epsilon[i] * l->n / s;
      if (l->w != NULL)
	m *= l->w[i];
      if (m == 0){
	/* Weight of this point is zero; do not include it in next iteration */
	k++;
      }
      else{
	/* Compare weight to old weight and update */
	if (fabs(m - lo->w[i - k]) > _IPLOESS_ROBUSTSTOP)
	  /* This weight has not converged yet */
	  c = 1;
	lo->w[i - k] = m;
	/* Convergence not attained if there is a change in the set of
	 * points with non-zero weight.
	 */ 
	if (lo->x[i - k] != l->x[i]){
	  lo->x[i - k] = l->x[i];
	  c = 1;
	}
	lo->y[i - k] = l->y[i];
      }
    }
    /* Convergence not attained if there is a change in the set of
     * points with non-zero weight.
     */
    if (lo->n != l->n - k){
      lo->n = l->n - k;
      c = 1;
    }
    lo->q = compute_q(lo->n, lo->alpha, lo->lambda);
    /* Another iteration older... */
    j++;
  } while (c && (j < _IPLOESS_ROBUSTMAXITERATIONS));

  /* Compute an approximate Loess fit from final robustified Loess fit */
  ipl = iploess(lo, n, l->x[0], l->x[l->n - 1]);

  /* Clean up */
  wfree(epsilon);
  loess_free(lo);

  return ipl;
}

/* iploess_evaluate(ipl, x): evaluates the approximate Loess fit ipl at x */
double iploess_evaluate(IPLoess *ipl, double x)
{
  Polynomial *p;
  int i;
  double r;

  /* Find index of interval x belongs to */
  if (x < ipl->left)
    i = 0;
  else if (x > ipl->right)
    i = ipl->n + 1;
  else{
    i = 1 + ipl->n * (x - ipl->left) / (ipl->right - ipl->left);
    if (i > ipl->n)
      i = ipl->n;
  }

  /* Initialise p to be the polynomial approximating the Loess fit in
   * this interval, and evaluate p at x.
   */
  p = polynomial_new(ipl->lambda + 1, ipl->p[i]);
  r = polynomial_evaluate(p, x);

  /* Clean up */
  polynomial_free(p);

  return r;
}

/* min(a, b): returns the smaller of a and b */
static double min(double a, double b)
{
  return (a < b ? a : b);
}

/* inc_interval(i, x, ipl, b, l): increments the interval index i and
 * updates x to the right end-point of the new interval, or b if b is
 * smaller of we reach the end of the intervals of ipl.
 */
static void inc_interval(int *i, double *x, IPLoess *ipl, double b, double l)
{
  if (++(*i) > ipl->n)
    /* We have moved to the right of the last interval */
    *x = b;
  else
    *x = min(b, ipl->left + l * (*i));
}

/* iploess_isd(ipl1, ipl2, a, b): computes the integral of the
 * difference between ipl1 and ipl2 squared in the interval from a to
 * b.
 */
double iploess_isd(IPLoess *ipl1, IPLoess *ipl2, double a, double b)
{
  double aggregate = 0, tmp, x, y1, y2, l1, l2;
  int negate = 0, i1, i2;
  Polynomial *p1, *p2;

  if (b < a){
    /* If b is less than a, swap a and b and remember that the
     * integral has to be negated.
     */
    tmp = a;
    a = b;
    b = tmp;
    negate = 1;
  }

  /* Determine first relevant interval of each interpolated loess fit
   * and set up intial polynomials.
   */
  x = a;
  l1 = (ipl1->right - ipl1->left) / ipl1->n;
  if (a < ipl1->left){
    i1 = 0;
    y1 = min(b, ipl1->left);
  }
  else if (a > ipl1->right){
    i1 = ipl1->n + 1;
    y1 = b;
  }
  else{
    i1 = 1 + (a - ipl1->left) / l1;
    if (i1 > ipl1->n)
      i1 = ipl1->n;
    y1 = min(b, ipl1->left + l1 * i1);
  }
  p1 = polynomial_new(ipl1->lambda + 1, ipl1->p[i1]);
  l2 = (ipl2->right - ipl2->left) / ipl2->n;
  if (a < ipl2->left){
    i2 = 0;
    y2 = min(b, ipl2->left);
  }
  else if (a > ipl2->right){
    i2 = ipl2->n + 1;
    y2 = b;
  }
  else{
    i2 = 1 + (a - ipl2->left) / l2;
    if (i2 > ipl2->n)
      i2 = ipl2->n;
    y2 = min(b, ipl2->left + l2 * i2);
  }
  p2 = polynomial_new(ipl2->lambda + 1, ipl2->p[i1]);

  /* Compute the integral by dividing the interval between a and b
   * into subintervals for which both interpolated loess fits have
   * fixed polynomials.
   */
  do{
    /* Find right end-point of current integral, compute interval's
     * contribution to the integral, and update intervals.
     */
    if (y1 < y2){
      aggregate += polynomial_isd(p1, p2, x, y1);
      x = y1;
      inc_interval(&i1, &y1, ipl1, b, l1);
      polynomial_free(p1);
      p1 = polynomial_new(ipl1->lambda + 1, ipl1->p[i1]);
    }
    else if (y2 < y1){
      aggregate += polynomial_isd(p1, p2, x, y2);
      x = y2;
      inc_interval(&i2, &y2, ipl2, b, l2);
      polynomial_free(p2);
      p2 = polynomial_new(ipl2->lambda + 1, ipl2->p[i1]);
    }
    else{
      aggregate += polynomial_isd(p1, p2, x, y1);
      x = y1;
      inc_interval(&i1, &y1, ipl1, b, l1);
      inc_interval(&i2, &y2, ipl2, b, l2);
      polynomial_free(p1);
      p1 = polynomial_new(ipl1->lambda + 1, ipl1->p[i1]);
      polynomial_free(p2);
      p2 = polynomial_new(ipl2->lambda + 1, ipl2->p[i1]);
    }
  } while (x < b);

  /* Clean up */
  polynomial_free(p1);
  polynomial_free(p2);

  if (negate)
    return -aggregate;
  else
    return aggregate;
}

/* iploess_free(l): frees memory used by the iploess structure
 * referenced by l.
 */
void iploess_free(IPLoess *ipl)
{
  int i;

  for (i = 0; i < ipl->n + 2; i++)
    wfree(ipl->p[i]);
  wfree(ipl->p);
  wfree(ipl);
}
