/* istvan.c: vanilla flavoured function for normalising expression data */
/* Rune Lyngs, March 2004 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include "istvan.h"
#include "llist.h"
#include "splay.h"
#include "rankselect.h"
#include "polynomial.h"
#include "loess.h"
#include "iploess.h"
#include "iterated_global_normalisation.h"
#include "absolute_rank_selection.h"
#include "relative_rank_selection.h"
#include "iterated_ars.h"
#include "reliability.h"
#include "wrappers.h"
#include "gmp_interface.h"

/* Global variables needed for passing parameters to functions
 * computing new normalising functions.
 */
static Function function = POLYNOMIAL;
static int robustify = 0, am = 0, intervals = 0;
static double alpha = 0.4, lowfraction, highfraction, scale;
static unsigned int degree = 4;
static double min, max;

/* xmalloc(n): Wrapper function for malloc checking return value */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "istvan");
#else
  return wmalloc(n);
#endif
}

/* amtransform(n, x, y): Transform data in x and y arrays to
 * mean/ratio space.
 */
static void amtransform(int n, double *x, double *y)
{
  double tmp;
  int i;
  for (i = 0; i < n; i++){
    tmp = x[i] + y[i];
    y[i] = x[i] - y[i];
    x[i] = tmp / 2;
  }
}

/* set_inference(n, x, y): This is the function that should be used
 * for inferring the normalising function from non-weighted invariant
 * sets.
 */
static void *set_inference(int n, double *x, double *y)
{
  Loess *l;
  IPLoess *ipl;

  if (am) amtransform(n, x, y);
  switch(function){
  case LOESS:
    l = loess(n, x, y, alpha, degree);
    /* If intervals is zero, use standard loess, otherwise use
     * interpolated loess.
     */
    if (intervals == 0){
      if (robustify)
	return loess_robustify(l);
      else
	return l;
    }
    else{
      if (robustify)
	ipl = iploess_robust(l, intervals, min, max);
      else
	ipl = iploess(l, intervals, min, max);
      loess_free(l);
      return ipl;
    }
    break;
  default:
  case POLYNOMIAL:
    return polynomial_lsdpolynomial(degree + 1, n, x, y, _ISTVAN_SMOOTHWEIGHT,
				    min, max);
    break;
  }
}

/* weighted_and_transformed_inference(n, x, y): This is the function
 * that should be used for inferring the normalising function from
 * weighted invariant sets that have been transformed.
 */
static void *weighted_and_transformed_inference(int n, double *x, double *y,
						double *w)
{
  Loess *l;
  IPLoess *ipl;

  switch(function){
  case LOESS:
    l = wloess(n, x, y, w, alpha, degree);
    /* If intervals is zero, use standard loess, otherwise use
     * interpolated loess.
     */
    if (intervals == 0){
      if (robustify)
	return loess_robustify(l);
      else
	return l;
    }
    else{
      if (robustify)
	ipl = iploess_robust(l, intervals, min, max);
      else
	ipl = iploess(l, intervals, min, max);
      loess_free(l);
      return ipl;
    }
    break;
  default:
  case POLYNOMIAL:
    return polynomial_wlsdpolynomial(degree + 1, n, x, y, w,
				     _ISTVAN_SMOOTHWEIGHT, min, max);
    exit(0);
    break;
  }
}

/* weighted_inference(n, x, y): This is the function that should be
 * used for inferring the normalising function from weighted invariant
 * sets.
 */
static void *weighted_inference(int n, double *x, double *y, double *w)
{
  /* Transform data to mean/ratio space if so instructed */
  if (am) amtransform(n, x, y);

  /* Infer normalising function */
  return weighted_and_transformed_inference(n, x, y, w);
}

/* eval_function(f, x): Compute f(x) */
static double eval_function(void *f, double x)
{
  switch(function){
  case LOESS:
    if (intervals == 0)
      return loess_evaluate((Loess *)f, x);
    else
      return iploess_evaluate((IPLoess *)f, x);
    break;
  default:
  case POLYNOMIAL:
    return polynomial_evaluate((Polynomial *)f, x);
    break;
  }
}

/* cmp_function(f, g): Determine whether f is (almost) equal to g */
static int cmp_function(void *f, void *g)
{
  switch(function){
  case LOESS:
    if (intervals == 0)
      return (loess_nisd((Loess *)f, (Loess *)g, min, max,
			 _ISTVAN_LOESSINTERVALS)
	      < _ISTVAN_CONVERGENCETHRESHOLD);
    else
      return (iploess_isd((IPLoess *)f, (IPLoess *)g, min, max)
	      < _ISTVAN_CONVERGENCETHRESHOLD);
    break;
  default:
  case POLYNOMIAL:
    return (polynomial_isd((Polynomial *)f, (Polynomial *)g, min, max) <
	    _ISTVAN_CONVERGENCETHRESHOLD);
    break;
  }
}

/* free_function(f): Deallocate memory used by function f */
static void free_function(void *f)
{
  switch(function){
  case LOESS:
    if (intervals == 0)
      loess_free((Loess *)f);
    else
      iploess_free((IPLoess *)f);
    break;
  default:
  case POLYNOMIAL:
    polynomial_free((Polynomial *)f);
    break;
  }
}

/* square(x): Compute x squared */
static double square(double x)
{
  return x * x;
}

/* bisquare(x, m, s, a): Compute the bisquare weight as defined for
 * loess fits, i.e. (1 - (x / 6 * m)^2)^2.
 */
static double bisquare(double x, double m, double s, double a)
{
  return square(1 - square(x / (6 * m)));
}

/* thresholdfunction(x, y, minx, maxx, miny, maxy, rx, ry, n):
 * Determine whether the point should be retained in a particular
 * round of iterated absolute rank selection.
 */
static int thresholdfunction(double x, double y, double minx, double maxx,
			     double miny, double maxy, int rx, int ry, int n)
{
  /* The threshold is obtained by a linear interpolation between
   * lowfraction and highfraction based on the average expression
   * level. t denotes the relative distance to the average minimum
   * expression level compared to the average maximum expression
   * level.
   */
  double t = (x + y - minx - miny) / (maxx + maxy - minx - miny);

  return (abs(rx - ry) <= n * (t * highfraction + (1 - t) * lowfraction));
}

/* sigmoid(n, weights): Transform the n numbers in the array weights
 * by passing them through a sigmoid function centered around the
 * average.
 */
static void sigmoid(int n, mp_type *weights)
{
  int i;
  mp_type avg, tmp, v;
  double sd;

  /* Compute average and sample standard deviation of weights */
  gmpi_init_set_ui(avg, 0);
  for (i = 0; i < n; i++)
    gmpi_add(avg, avg, weights[i]);
  gmpi_init_set_ui(tmp, n);
  gmpi_div(avg, avg, tmp);
  gmpi_init_set_ui(v, 0);
  for (i = 0; i < n; i++){
    /* Remember the difference between weight and average as we will
     * need it later.
     */
    gmpi_sub(weights[i], weights[i], avg);
    gmpi_mul(tmp, weights[i], weights[i]);
    gmpi_add(v, v, tmp);
  }
  if (n > 1){
    gmpi_set_ui(tmp, n - 1);
    gmpi_div(v, v, tmp);
  }
  sd = sqrt(gmpi_get_d(v));
  gmpi_clear(v);
  /* Make sure that sd is not zero */
  if (sd == 0){
    sd = gmpi_get_d(avg);
    if (sd == 0)
      sd = 1;
  }

  /* Subject each weight w to the sigmoid function
   * w = 1 / (1 + e^(-scale * (w - a) / s)) where a is the average
   * weight and s is the sample standard deviation.
   */
  for (i = 0; i < n; i++)
    gmpi_set_d(weights[i], 1 / (1 + exp(scale * gmpi_get_d(weights[i]) / sd)));

  /* Clean up */
  gmpi_clear(avg);
  gmpi_clear(tmp);
}

/* Return whether x < y */
static int less_than(double *x, double *y)
{
  return (*x < *y);
}

/* print_usage(fp, name): prints a short description of how to use the
 * program.
 */
static void print_usage(FILE *fp, char *name)
{
  if (name == NULL)
    name = "istvan";

  char *usage_txt = "\nUsage: %s [options] [infile [outfile]]\n\n%s is a command line program for normalising a pair of expression array data\nsets, i.e. to make the expression levels directly comparable. It implements\nseveral invariant set methods - global, global with iterative\nrobustification, rank invariance selection, iterated rank invariance\nselection, order invariance selection, and reliability - and two classes of\nnormalisation functions that can be fitted - loess fits and fixed degre\npolynomials. The raw data is read from infile, or standard input if no infile\nis specified, and the normalised data is written to outfile, or standard\noutput if no outfile is specified. Legal options (options followed by a\ncapital letter require an argument, options followed by a capital letter in\nsquare brackets take an optional argument) are:\n\n -h -H -?: Print this information and stop\n -a: Normalise in A/M transformed space\n -c: Normalising in A/M transformed space usually results in modifications of\n     measurements for both samples. The -c option changes this behavious such\n     that only the measurements of the second sample are modified. If this\n      normalisation is part of a multi sample experiment where all samples\n     are normalised to a central reference sample, use the -c option to\n     specify that the first sample is the reference sample. This option does\n     not set A/M transformed normalissation, so it needs to be used in\n     combination with the -a option\n -e: Do not work in log transformed space; default behaviour is to replace\n     all measurements with their (base e) logarithm\n -xN: Exclude the N genes with lowest average intensity and the N genes with\n      highest average intensity (average taken in log space unless the -e\n      option is specified) before inferring normalisation function\n -g: Global normalisation, i.e. use all genes with equal weight for\n     normalisation function inference\n -b[X]: Use iterated global normalisation to determine gene weights for\n        normalisation function inference; the new weight is the bisquare\n        function of the distance to the previous estimate relative to X times\n        the median distance, where the default value of X is 6\n -t[X]: Use iterated global normalisation to determine gene weights for\n        normalisation function inference; the new weight is the Gaussian\n        function of the distance to the previous estimate relative to X times\n        the sample variance, where the default value of X is 1\n -r[N]: Use invariant rank selection with threshold N to compute an invariant\n        set; the default value of N is 15\n -i[F]: Use Iterated invariant rank selection to compute an invariant set; in\n        each iteration a gene is retained if its rank difference is at\n        most F times the number of remaining genes; the default value of\n        F is 0.02\n -jF: Set the threshold percentage for iterated invariant rank selection to be\n      an interpolation between the rank specified by the -i option (or the\n      default value) at low expression values and F at high expression\n      values; this option does not set the invariant set method so it needs\n      to be used in combination with the -i option\n -o[N]: Use invariant order selection with threshold N to compute an\n        invariant set; the default value of N is 15\n -q[T]: Use reliability with temperature T to compute gene weights for\n        normalisation function inference; the default value of T is 1.5\n -s[X]: Use reliability with reliabilities measured in terms of the distance\n        to the average reliability multiplied by X divided by the sample\n        standard deviation passed through a sigmoid function; the default\n        value of X is 1\n -p[D]: Use polynomials of degree D as class of normalising functions; this\n        is the default class of functions with the default value of D being 4\n -l[D]: Use loess fits of order D as class of normalising function; the\n        default value of D is 1\n -nA: Set the neighbourhood parameter of loess fits to A; the default value\n      of A is 0.4\n -m: Use robust loess fits as class of normalising functions; to change the\n     order from the default value of 1 use the -l option\n -k[N]: Approximate loess fit with N piecewise polynomials of the same order;\n        the default value of N is the number of measurements divided by 100;\n        this option does not set the type of normalisation function, so it\n        needs to be used in combination with the -l (and possibly the -m)\n        option\n -fN: write output to file named N; this option is useful for sending the\n      normalised data to a file while still reading the unnormalised data\n      from standard input\n -d: Read (and write) the data as a pair of lists; default is as a list of\n     pairs\n";

  fprintf(fp, usage_txt, name, name);
}

int main(int argc, char **argv)
{
  LList *data;
  int i, j, k, l, n, logtransform = 1, threshold, list_of_pairs = 1,
    reference = 0;
  unsigned int exclude = 0;
  double temperature, tmp, *item, *x, *y, *oldx, *oldy;
  void *inferred_function = NULL;
  Method method = GLOBAL;
  FILE *input = stdin, *output = stdout;

  /* Analyse command line options */
  opterr = 0;
  while ((i = getopt(argc, argv,
		     "ab::cdef:ghi::j:k::l::mn:o::p::q::r::s::t::x:")) >= 0){
    switch(i){
    case 'h':
    case '?':
      print_usage(stdout, argv[0]);
      exit(0);
      break;
    case 'a':
      am = 1;
      break;
    case 'b':
      method = IGNBISQUARE;
      scale = 6;
      if (optarg && ((sscanf(optarg, "%lf", &scale) == 0) || (scale <= 0))){
	fprintf
	  (stderr,
	   "Could not parse %s as scaling parameter for bisquare function\n",
	   optarg);
	exit(1);
      }
      break;
    case 'c':
      reference = 1;
      break;
    case 'd':
      list_of_pairs = 0;
      break;
    case 'e':
      logtransform = 0;
      break;
    case 'f':
      if ((output = fopen(optarg, "w")) == NULL){
	fprintf(stderr, "Could not open file %s for output\n", optarg);
	exit(1);
      }
      break;
    case 'g':
      method = GLOBAL;
      break;
    case 'i':
      method = ITERATEDRANKSELECTION;
      lowfraction = 0.02;
      if (highfraction < 0)
	highfraction = lowfraction;
      if (optarg && ((sscanf(optarg, "%lf", &lowfraction) == 0)
		     || (lowfraction < 0) || (lowfraction > 1))){
	fprintf(stderr, "Could not parse %s as a fraction\n", optarg);
	exit(1);
      }
      break;
    case 'j':
      if ((sscanf(optarg, "%lf", &highfraction) == 0) || (highfraction < 0)
	  || (highfraction > 1)){
	fprintf(stderr, "Could not parse %s as a fraction\n", optarg);
	exit(1);
      }
      break;
    case 'k':
      intervals = -1;
      if (optarg
	  && ((sscanf(optarg, "%u", &intervals) == 0) || intervals <= 0)){
	fprintf(stderr, "Could not parse %s as a threshold\n", optarg);
	exit(1);
      }
      break;
    case 'l':
      function = LOESS;
      degree = 1;
      if (optarg && (sscanf(optarg, "%u", &degree) == 0)){
	fprintf(stderr, "Could not parse %s as loess lambda parameter\n",
		optarg);
	exit(1);
      }
      break;
    case 'm':
      robustify = 1;
      break;
    case 'n':
      if (sscanf(optarg, "%lf", &alpha) == 0){
	fprintf(stderr,
		"Could not parse %s as loess neighbourhood parameter\n",
		optarg);
	exit(1);
      }
      break;
    case 'o':
      method = ORDERSELECTION;
      threshold = 15;
      if (optarg && (sscanf(optarg, "%u", &threshold) == 0)){
	fprintf(stderr, "Could not parse %s as a threshold\n", optarg);
	exit(1);
      }
      break;
    case 'p':
      function = POLYNOMIAL;
      degree = 4;
      if (optarg && (sscanf(optarg, "%u", &degree) == 0)){
	fprintf(stderr, "Could not parse %s as order of a polynomial\n",
		optarg);
	exit(1);
      }
      break;
    case 'q':
      if (method != SIGMOIDEDRELIABILITY)
	method = RELIABILITY;
      temperature = 1.5;
      if (optarg && (sscanf(optarg, "%lf", &temperature) == 0)){
	fprintf(stderr,
		"Could not parse %s as a partition function temperature\n",
		optarg);
	exit(1);
      }
      break;
    case 'r':
      method = RANKSELECTION;
      threshold = 15;
      if (optarg && (sscanf(optarg, "%u", &threshold) == 0)){
	fprintf(stderr, "Could not parse %s as a threshold\n", optarg);
	exit(1);
      }
      break;
    case 's':
      method = SIGMOIDEDRELIABILITY;
      scale = 4;
      if (optarg && ((sscanf(optarg, "%lf", &scale) == 0) || (scale < 0))){
	fprintf(stderr, "Could not parse %s as a slope parameter for a sigmoid function\n", optarg);
	exit(1);
      }
      break;
    case 't':
      method = IGNNORMAL;
      scale = 1;
      if (optarg && ((sscanf(optarg, "%lf", &scale) == 0) || (scale <= 0))){
	fprintf(stderr, "Could not parse %s as scaling parameter for weight function\n", optarg);
	exit(1);
      }
      break;
    case 'x':
      if (sscanf(optarg, "%u", &exclude) == 0){
	fprintf(stderr,
		"Could not parse %s as number of extreme genes to exclude\n",
		optarg);
	exit(1);
      }
      break;
    default:
      /* Unrecognised option */
      fprintf(stderr, "Unrecognised option: %c\n", i);
      print_usage(stderr, argv[0]);
      exit(1);
    }
  }

  if (optind < argc){
    /* Input file specified */
    if ((input = fopen(argv[optind], "r")) == NULL){
      fprintf(stderr, "Could not open %s for reading\n", argv[optind]);
      exit(1);
    }
    if (optind + 1 < argc){
      /* Output file specified */
      if ((output = fopen(argv[optind + 1], "w")) == NULL){
	fprintf(stderr, "Could not open %s for reading\n", argv[optind + 1]);
	exit(1);
      }
    }
  }

  /* Read data from input and temporarily store it in an LList */
  data = MakeLList();
  while (fscanf(input, "%lf", &tmp) == 1){
    item = (double *)xmalloc(sizeof(double));
    *item = tmp;
    Enqueue(data, item);
  }
  if (input != stdin)
    fclose(input);

  if (Length(data) % 2 != 0){
    fprintf(stderr, "The data contains an odd number (%d) of measurements; I will continue,\ndiscarding the last measurement read, but probably something is wrong.\n", Length(data));
    wfree(Dequeue(data));
  }

  /* Move data to an array and convert to log space unless told not to do so */
  n = Length(data) / 2;
  x = (double *)xmalloc(Length(data) * sizeof(double));

  for (i = 0; Length(data) > 0; i++){
    item = Pop(data);
    if (logtransform)
      *item = log(*item);
    /* If the data layout is as a list of pairs of measurements, every
     * other measurement should be placed in the second half of the
     * array; if the data layout is as a pair of lists of measurement,
     * the measurement should be placed in the array in the same order
     * as it was read.
     */
    x[i / (1 + list_of_pairs) + n * ((i + 1) % 2) * list_of_pairs] = *item;
    wfree(item);
  }
  DestroyLList(data, NULL);
  /* The measurements of the first experiment should now be located in
   * the second half of the array.
   */
  y = x + n;

  l = n;
  if (exclude > 0){
    /* Move the exclude genes with highest and the exclude genes with
     * lowest average intensity to the end of the array so that we can
     * leave them out of the normalisation function inference.
     */
    if (2 * exclude >= n)
      /* Hmm, excluding this many genes would leave nothing left */
      fprintf(stderr, "Excluding the %u genes with highest intensity and the %u genes with lowest\nintensity would leave no genes remaining; ignoring specified exclusion\n", exclude, exclude);
    else{
      /* Store old values and allocate new array */
      oldx = x;
      oldy = y;
      x = (double *)xmalloc(2 * (n - 2 * exclude) * sizeof(double));
      y = x + (n - 2 * exclude);
      /* Compute summed intensities */
      item = (double *)xmalloc(n * sizeof(double));
      for (i = 0; i < n; i++)
	item[i] = oldx[i] + oldy[i];
      /* Determine exclude'th smallest and highest summed intensities */
      j = rankselect(item, exclude - 1, n, sizeof(double),
		     (int (*)(void *, void *))less_than);
      k = rankselect(item, n - exclude, n, sizeof(double),
		     (int (*)(void *, void *))less_than);
      /* Move genes inside this range to new arrays */
      l = 0;
      for (i = 0; i < n; i++)
	if ((item[i] > item[j]) && (item[i] < item[k])){
	  x[l] = oldx[i];
	  y[l] = oldy[i];
	  l++;
	}
      /* Did we move enough genes? */
      i = 0;
      while (l < n - 2 * exclude)
	if ((item[i] == item[j]) || (item[i] == item[k])){
	  x[l] = oldx[i];
	  y[l] = oldy[i];
	  l++;
	}
	else
	  i++;
      /* Clean up */
      wfree(item);
    }
  }
      
  /* Update number of intervals for interpolated loess if default
   * value was specified.
   */
  if (intervals < 0) intervals = (n < 200 ? 1 : n / 100);

  /* Find smallest and largest first coordinate */
  min = x[0];
  max = x[0];
  for (i = 1; i < n; i++){
    if (x[i] > max) max = x[i];
    if (x[i] < min) min = x[i];
  }

  /* Infer normalisation function as specified */
  switch(method){
  case IGNBISQUARE:
    /* Convert data to mean/ratio space if so instructed */
    if (am) amtransform(l, x, y);
    inferred_function = iterated_global_normalisation
      (l, x, y, (_IGN_newfunction)weighted_and_transformed_inference,
       (_IGN_evalfunction)eval_function, (_IGN_cmpfunction)cmp_function,
       (_IGN_freefunction)free_function, (_IGN_weight)bisquare);
    break;
  case GLOBAL:
    /* The arrays x and y are passed on to the set_inference function
     * where they will be transformed to A/M space if so specified.
     */
    inferred_function = set_inference(l, x, y);
    break;
  case ITERATEDRANKSELECTION:
    inferred_function = iterated_ars(l, x, y, (_IARS_newfunction)set_inference,
				     (_IARS_threshold)thresholdfunction);
    /* Rank selection has to be done on untransformed data, but we
     * still need to transform the data before applying the
     * normalisation.
     */
    if (am && (l == n)) amtransform(n, x, y);
    break;
  case ORDERSELECTION:
    inferred_function = relative_rank_selection
      (l, x, y, threshold, (_RRS_newfunction)set_inference);
    /* Order selection has to be done on untransformed data, but we
     * still need to transform the data before applying the
     * normalisation.
     */
    if (am && (l == n)) amtransform(n, x, y);
    break;
  case RELIABILITY:
    /* The arrays x and y are passed on to the weighted_inference
     * function where they will be transformed to A/M space if so
     * specified.
     */
    inferred_function = reliability
      (l, x, y, temperature, (_RELIABILITY_newfunction)weighted_inference,
       NULL);
    break;
  case RANKSELECTION:
    inferred_function = absolute_rank_selection
      (l, x, y, threshold, (_ARS_newfunction)set_inference);
    /* Rank selection has to be done on untransformed data, but we
     * still need to transform the data before applying the
     * normalisation.
     */
    if (am && (l == n)) amtransform(n, x, y);
    break;
  case SIGMOIDEDRELIABILITY:
    /* The arrays x and y are passed on to the weighted_inference
     * function where they will be transformed to A/M space if so
     * specified.
     */
    inferred_function = reliability
      (l, x, y, temperature, (_RELIABILITY_newfunction)weighted_inference,
       (_RELIABILITY_transform)sigmoid);
    break;
  case IGNNORMAL:
    /* Convert data to mean/ratio space if so instructed */
    if (am) amtransform(l, x, y);
    inferred_function = iterated_global_normalisation
      (l, x, y, (_IGN_newfunction)weighted_and_transformed_inference,
       (_IGN_evalfunction)eval_function, (_IGN_cmpfunction)cmp_function,
       (_IGN_freefunction)free_function, NULL);
    break;
  }

  /* If some genes were excluded, recover the full set */
  if ((exclude > 0) && (2 * exclude < n)){
    wfree(x);
    x = oldx;
    y = oldy;
    if (am) amtransform(n, x, y);
  }

  /* Normalise data according to inferred function; if data was A/M
   * transformed, also transform data back to experiment space.
   */
  if (am)
    for (i = 0; i < n; i++){
      if (!reference)
	/* Apply normalisation now, such that it affects both samples */
	y[i] = y[i] - eval_function(inferred_function, x[i]);
      tmp = x[i] - y[i] / 2;
      x[i] = x[i] + y[i] / 2;
      if (reference)
	/* Apply normalisation modification to second sample */
	x[i] -= eval_function(inferred_function, x[i]);
      y[i] = tmp;
#ifdef FUBAR
      tmp = x[i] + y[i] / 2;
      y[i] = x[i] - y[i] / 2;
      if (reference)
	/* Apply normalisation modification to second sample */
	y[i] += eval_function(inferred_function, x[i]);
      x[i] = tmp;
#endif
    }
  else
    for (i = 0; i < n; i++)
      x[i] = eval_function(inferred_function, x[i]);

  /* Transform back to experiment space if log transformation has been
   * applied.
   */
  if (logtransform)
    for (i = 0; i < n; i++){
      x[i] = exp(x[i]);
      y[i] = exp(y[i]);
    }

  /* Output normalised data */
  if (list_of_pairs)
    for (i = 0; i < n; i++)
      fprintf(output, "%lf %lf\n", y[i], x[i]);
  else{
    fprintf(output, "%lf", y[0]);
    for (i = 1; i < n; i++)
      fprintf(output, " %lf", y[i]);
    fprintf(output, "\n%lf", x[0]);
    for (i = 1; i < n; i++)
      fprintf(output, " %lf", x[i]);
  }

  /* Clean up */
  if (output != stdout)
    fclose(output);
  free_function(inferred_function);
  wfree(x);
  splay_freeunusednodes();
  FreeUnusedLListNodes();

#ifdef MPROFILING
  printf("Undeallocated memory identifications: ");
  print_allocated();
#endif

  return 0;
}
