/* iterated_ars.c: Implementations of functions for computing
 * normalising functions based on iterated absolute rank selection
 * invariant sets. The threshold for a particular point can depend on
 * its expression level in the two experiments, the minimum and
 * maximum expression level in the two experiments of any remaining
 * points, its rank among the remaining points in the two experiments
 * and the number of points remaining.
 */
/* Rune Lyngs, February 2003 */
#include <stdio.h>
#include <stdlib.h>
#include "iterated_ars.h"
#include "mergesort.h"
#include "wrappers.h"

/* xmalloc(n): wrapper function for malloc, testing whether memory was
 * correctly allocated.
 */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "iterated_ars");
#else
  return wmalloc(n);
#endif
}

/* cmp_idpair(a, b): return true if value of a is less than value of b */
static int cmp_idpair(_IARS_idpair *a, _IARS_idpair *b)
{
  return (a->value < b->value);
}

/* iterated_ars(n, x, y, new_function, threshold): computes the
 * normalising function based on the points selected by iterated
 * absolute rank selection from the n points given by x and y. The
 * points selected are passed to new_function to obtain the
 * normalising function, and in each iteration information about each
 * individual point is passed to threshold that should return TRUE if
 * the point should be retained and FALSE otherwise.
 */
void *iterated_ars(int n, double *x, double *y, _IARS_newfunction new_function,
		   _IARS_threshold threshold)
{
  _IARS_idpair *sortx = (_IARS_idpair *)xmalloc(n * sizeof(_IARS_idpair)),
    *sorty = (_IARS_idpair *)xmalloc(n * sizeof(_IARS_idpair));
  int *rx = (int *)xmalloc(n * sizeof(int)),
    *ry = (int *)xmalloc(n * sizeof(int));
  double *xs, *ys;
  int i, j, k, m, stop = _IARS_MINIMUM_FRACTION * n;
  int *d;
  void *f;

  /* Sort elements according to x and y values, respectively */
  for (i = 0; i < n; i++){
    sortx[i].index = i;
    sorty[i].index = i;
    sortx[i].value = x[i];
    sorty[i].value = y[i];
  }
  mergesort((void *)sortx, n, sizeof(_IARS_idpair),
	    (int (*)(void *, void *))cmp_idpair);
  mergesort((void *)sorty, n, sizeof(_IARS_idpair),
	    (int (*)(void *, void *))cmp_idpair);

  /* Determine element rank according to x and y values, respectively */
  for (i = 0; i < n; i++){
    rx[sortx[i].index] = i;
    ry[sorty[i].index] = i;
  }

  m = n;
  for (;;){
    /* Eliminate points with too big rank difference */
    j = 0;
    for (i = 0; i < m; i++){
      if (threshold(sortx[i].value, y[sortx[i].index], sortx[0].value,
		    sortx[m - 1].value, sorty[0].value, sorty[m - 1].value,
		    rx[sortx[i].index], ry[sortx[i].index], m)){
	/* Retain point */
	sortx[j].index = sortx[i].index;
	sortx[j].value = sortx[i].value;
	j++;
      }
      else
	/* Eliminate point */
	rx[sortx[i].index] = -rx[sortx[i].index] - 1;
    }

    /* Stop point elimination if current iteration would eliminate too
     * many of the remaining points, or if current iteration does not
     * eliminate any points.
     */
    if (j == m)
      break;
    if (j < stop){
      /* Find smallest rank difference such that the set of all points
       * with at most that rank difference is of size at least
       * stop.
       */
      /* Start by restoring ranks of points destined for elimination */
      for (i = 0; i < m; i++)
	if (rx[sorty[i].index] < 0)
	  rx[sorty[i].index] = -rx[sorty[i].index] + 1;
      /* Find largest rank difference for remaining points */
      j = 0;
      for (i = 0; i < m; i++)
	if (abs(ry[sorty[i].index] - rx[sorty[i].index]) > j)
	  j = abs(ry[sorty[i].index] - rx[sorty[i].index]);
      /* Count sort the rank differences */
      d = (int *)xmalloc((j + 1) * sizeof(int));
      for (i = 0; i < j; i++)
	d[i] = 0;
      for (i = 0; i < m; i++)
	d[abs(ry[sorty[i].index] - rx[sorty[i].index])]++;
      /* Find smallest i such that there are at least stop points with
       * rank difference at most i.
       */
      for (i = 0; d[i] < stop; i++)
	d[i + 1] += d[i];
      wfree(d);
      /* Move points with rank difference at most i to front of sorty */
      k = 0;
      for (j = 0; j < m; j++)
	if (abs(ry[sorty[j].index] - abs(rx[sorty[j].index])) <= i){
	  sorty[k].index = sorty[j].index;
	  k++;
	}
      m = k;
#ifdef _IARS_FORCEDSTOPWARNING
      fprintf(stderr, "Further iteration would lead to set with fewer than %d points.\nForced stop resulted in set with %d points.\n", stop, n);
#endif
      break;
    }

    /* Update number of points remaining */
    m = j;

    /* We still need to contract sorty to only contain retained
     * points, and update rx and ry to reflect the ranks in the
     * remaining set of points; eliminated points have negative rank
     * in rx.
     */
    j = 0;
    for (i = 0; i < m; i++){
      /* Update rx value */
      rx[sortx[i].index] = i;
      /* Ignore eliminated points */
      while (rx[sorty[j].index] < 0) j++;
      sorty[i].index = sorty[j].index;
      sorty[i].value = sorty[j].value;
      ry[sorty[i].index] = i;
      j++;
    }
  }

  if (new_function != NULL){
    /* Copy retained points to xs and ys */
    xs = (double *)xmalloc(m * sizeof(double));
    ys = (double *)xmalloc(m * sizeof(double));
    for (i = 0; i < m; i++){
      xs[i] = x[sorty[i].index];
      ys[i] = y[sorty[i].index];
    }

    /* Compute normalising function */
    f = new_function(m, xs, ys);

    /* Clean up */
    wfree(sortx);
    wfree(sorty);
    wfree(rx);
    wfree(ry);
    wfree(xs);
    wfree(ys);

    return f;
  }
  else{
    /* Create array with i'th element 1 if the i'th point was
     * retained, 0 otherwise
     */
    xs = (double *)xmalloc(n * sizeof(double));
    for (i = 0; i < n; i++)
      xs[i] = 0;
    for (i = 0; i < m; i++)
      xs[sorty[i].index] = 1;

    /* Clean up */
    wfree(sortx);
    wfree(sorty);
    wfree(rx);
    wfree(ry);

    return (void *)xs;
  }
}
