/*******************************************************************

    llist.c
  
    Implementation of a general purpose linked list with a stack and
    queue interface.
		
    Christian Storm (cstorm@daimi.au.dk), May 1995

    Revision History:

    January 2001. Modified LListMap to take a variable list of
                  arguments, instead of a void pointer; minor
		  change to InitCounter to allow the use of FIRST
		  and LAST.
    March 2003.   Modified DestroyLList and DestroyLLists to take a
                  function for deleting elements stored in an LList
		  as parameter.
		  Rune Lyngs (rlyngsoe@brics.dk)
    August 2003.  Added node recycling code.
                  Rune Lyngs (rlyngsoe@brics.dk)

********************************************************************/

#include <stdlib.h> 
#include <stdio.h>
#include <stdarg.h>
#include "llist.h"
#include "wrappers.h"

static LListNode *recycledllistnodes = NULL;

/* Wrapper for malloc; checks whether memory was available. */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "llist");
#else
  return wmalloc(n);
#endif
}

/* createnode(elm, prev, next, ref): Create a new llist node with elm
 * as elm, prev as predecessor, next as successor and referenced by
 * ref.
 */
static LListNode *createnode(void *elm, LListNode *prev, LListNode *next,
			     LListNode **ref)
{
  LListNode *new;

  /* Allocate memory for the new llist node if a recycled llist node is
   * not available.
   */
  if (recycledllistnodes == NULL)
    new = (LListNode *)xmalloc(sizeof(LListNode));
  else{
    new = recycledllistnodes;
    recycledllistnodes = recycledllistnodes->next;
  }

  /* Initialise llistnode */
  new->elm = elm;
  new->prev = prev;
  if (prev != NULL)
    prev->next = new;
  new->next = next;
  if (next != NULL)
    next->prev = new;
  if (ref != NULL)
    *ref = new;

  return new;
}

/* Initialize an empty list. */
LList *MakeLList()
{
  /* Allocate space and initialize a 'LList' */
  LList *llist = (LList *)xmalloc(sizeof(LList));
  llist->first = createnode(NULL, NULL, NULL, &(llist->first));
  llist->last = createnode(NULL, llist->first, NULL, &(llist->last));
  llist->count = 0;

  return llist;
}

/* Initialize an array of n empty lists. */
LList *MakeLLists(int n)
{
  int i;
  LList *llists;

  /* Allocate space and initialize all 'LList's */
  llists = (LList *)xmalloc(n * sizeof(LList));
  for (i = 0; i < n; i++){
    llists[i].first = createnode(NULL, NULL, NULL, &(llists[i].first));
    llists[i].last = createnode(NULL, llists[i].first, NULL,
				&(llists[i].last));
    llists[i].count = 0;
  }

  return llists;
}

/* splay_preallocatenodes(n): creates n new nodes that are
 * inserted in the list of recycled nodes. We have to create each node
 * individually instead of just allocating the nodes as one big chunk
 * of memory as we would otherwise run into problems when trying to
 * free individual nodes. So the only reason there might be for
 * preallocating nodes is to (hopefully) have the nodes located
 * consecutively in memory, thus perhaps yielding a better cache
 * performance.
 */
void PreallocateLListNodes(int n)
{
  int i;
  LListNode *new;

  for (i = 0; i < n; i++){
    new = (LListNode *)xmalloc(sizeof(LListNode));
    new->next = recycledllistnodes;
    recycledllistnodes = new;
  }
}

/* FreeUnusedLListNodes(): Frees memory of all nodes currently in the
 * list of recycled nodes.
 */
void FreeUnusedLListNodes()
{
  LListNode *tmp = recycledllistnodes;

  while (recycledllistnodes != NULL){
    tmp = recycledllistnodes->next;
    wfree(recycledllistnodes);
    recycledllistnodes = tmp;
  }
}

/* Destroys the LList-structure. If not NULL, the function destroy_elm
 * is used to destroy the elements stored in the list. The function
 * makes it very easy to lose all contact with these elements ... use
 * with care.
 */
void DestroyLList(LList *llist, void (*destroy_elm)(void *))
{
  LListNode *temp;

  /* free all the LListNodes */
  while ((temp = llist->first->next) != NULL)
    {
      llist->first->next = temp->next;
      if ((destroy_elm != NULL) && (temp != llist->last))
	destroy_elm(temp->elm);
      wfree(temp);
    }
  wfree(llist->first);

  /* free the LList */
  wfree(llist);
}

/* Destroys an array of n LList structures. If not NULL, the function
 * destroy_elm is used to destroy the elements stored in the list. The
 * function makes it very easy to lose all contact with these elements
 * ... use with care.
 */
void DestroyLLists(LList *llist, int n, void (*destroy_elm)(void *))
{
  LListNode *temp;
  int i;

  /* free all the LListNodes */
  for (i = 0; i < n; i++){
    while ((temp = llist[i].first->next) != NULL){
      llist[i].first->next = temp->next;
      if ((destroy_elm != NULL) && (temp != llist[i].last))
	destroy_elm(temp->elm);
      wfree(temp);
    }
    wfree(llist[i].first);
  }

  /* free the LList array */
  wfree(llist);
}

/* Deletes the LList-structure. If not NULL, the function delete_elm
 * is used to delete the elements stored in the list. The function
 * makes it very easy to lose all contact with these elements ... use
 * with care. Contrary to DestroyLList the nodes of the llist are
 * recycled and not freed.
 */
void DeleteLList(LList *llist, void (*delete_elm)(void *))
{
  LListNode *temp;

  /* free all the LListNodes */
  while ((temp = llist->first->next) != NULL)
    {
      llist->first->next = temp->next;
      if ((delete_elm != NULL) && (temp != llist->last))
	delete_elm(temp->elm);
      temp->next = recycledllistnodes;
      recycledllistnodes = temp;
    }
  llist->first->next = recycledllistnodes;
  recycledllistnodes = llist->first;

  /* free the LList */
  wfree(llist);
}

/* Deletes an array of n LList structures. If not NULL, the function
 * delete_elm is used to delete the elements stored in the list. The
 * function makes it very easy to lose all contact with these elements
 * ... use with care. Contrary to DestroyLLists the nodes of the
 * llists are recycled and not freed.
 */
void DeleteLLists(LList *llist, int n, void (*delete_elm)(void *))
{
  LListNode *temp;
  int i;

  /* free all the LListNodes */
  for (i = 0; i < n; i++){
    while ((temp = llist[i].first->next) != NULL){
      llist[i].first->next = temp->next;
      if ((delete_elm != NULL) && (temp != llist[i].last))
	delete_elm(temp->elm);
      temp->next = recycledllistnodes;
      recycledllistnodes = temp;
    }
    (llist[i].first)->next = recycledllistnodes;
    recycledllistnodes = llist[i].first;
  }

  /* free the LList array */
  wfree(llist);
}

/* Inserts a new node to the right of `lnode', thus it would be a BIG
 * mistake to call with `lnode' equal to `llist->last'.  
 */
void Insert(LList *llist, LListNode *lnode, void *elm)
{
  createnode(elm, lnode, lnode->next, NULL);
  (llist->count)++;
}

/* Removes the node `lnode'. It shall only be used the remove nodes
 * inserted using `Insert'.  
 */
void *Remove(LList *llist, LListNode *lnode)
{
  void *elm = lnode->elm;
  
  (lnode->prev)->next = lnode->next;
  (lnode->next)->prev = lnode->prev;
  (llist->count)--;
  lnode->next = recycledllistnodes;
  recycledllistnodes = lnode;

  return elm;
}

/* Functions implementing a stack- and queue-interface to the linked
 * list using the generic `Insert' and `Remove' functions.
 */
void Push(LList *llist, void *elm)
{
  Insert(llist, llist->first, elm);
}

void *Pop(LList *llist)
{
  void *elm = NULL;

  if ((llist->first)->next != llist->last)
    /* The list is not empty */
    elm = Remove(llist, (llist->first)->next);

  return elm;
}

void *Top(LList *llist)
{
  /* If list is empty we return the elm member of the terminal
   * `stopblock', which is NULL.
   */
  return llist->first->next->elm;
}

void Enqueue(LList *llist, void *elm)
{
  Insert(llist, (llist->last)->prev, elm);
}

void *Dequeue(LList *llist)
{
  void *elm = NULL;

  if ((llist->last)->prev != llist->first)
    /* The list is not empty */
    elm = Remove(llist, (llist->last)->prev);

  return elm;
}

void *Bottom(LList *llist)
{
  /* If list is empty we return the elm member of the initial
   * `stopblock', which is NULL.
   */
  return llist->last->prev->elm;
}

/* Functions used to iterate through a linked list. */

void LListMap(LList *llist, void (*f)(void *, va_list), ...)
{
  LListNode *current = llist->first;
  va_list args;

  while ((current = current->next) != llist->last){
    va_start(args, *f);
    (*f)(current->elm, args);
    va_end(args);
  }
}

LListCounter *MakeCounter(LList *llist, int pos)
{
  LListCounter *lcounter = (LListCounter *)xmalloc(sizeof(LListCounter));

  InitCounter(lcounter, llist, pos);
  
  return lcounter;
}

void InitCounter(LListCounter *lcounter, LList *llist, int pos)
{
  if (pos == LAST)
    pos = llist->count;

  if (pos <= llist->count - 1 - pos)
    {
      /* The node 'pos' is not closest to the last node */
      lcounter->pos = 0;
      lcounter->current = (llist->first)->next;
    }
  else
    {
      /* The node 'pos' is closest to the last node */
      lcounter->pos = llist->count - 1;
      lcounter->current = (llist->last)->prev;
    }
  SetCounter(lcounter, pos);
}

/* Positions are from -1 to the number of elements n in the llist; -1
 * sets the counter to the initial `stopblock', n sets the counter to
 * the terminal `stopblock', and a position i in the range from 0 to n
 * - 1 sets the counter to the i'th member of the llist (counting from
 * 0).
 */
void *SetCounter(LListCounter *lcounter, int pos)
{
  if (pos == LAST){
    /* Move all the way to the right */
      while ((lcounter->current)->next != NULL)
        {
          (lcounter->pos)++;
          lcounter->current = (lcounter->current)->next;
        }
    }
  else{
    if (pos < 0)
      pos = FIRST;
    if (pos > lcounter->pos)
      {
	/* Move 'pos - lcounter->pos' nodes to the right */
	while (lcounter->pos < pos && (lcounter->current)->next != NULL)
	  {
	    (lcounter->pos)++;
	    lcounter->current = (lcounter->current)->next;
	  }
      }
    else if (pos < lcounter->pos)
      {
	/* Move 'lcounter->pos - pos' nodes to the left */
	while (lcounter->pos > pos && (lcounter->current)->prev != NULL)
	  {
	    (lcounter->pos)--;
	    lcounter->current = (lcounter->current)->prev;
	  }
      }
  }

  return (lcounter->current)->elm;
}

void *Next(LListCounter *lcounter)
{
  return SetCounter(lcounter, lcounter->pos + 1);
}

void *Prev(LListCounter *lcounter)
{
  return SetCounter(lcounter, lcounter->pos - 1);
}

/* Removes the current node of a counter. It shall only be used when
 * the counter is not at FIRST or LAST. The counter is set to the
 * preceeding element.
 */
void *RemoveMoveLeft(LList *llist, LListCounter *lcounter)
{
  LListNode *tmp = lcounter->current;
  void *elm = tmp->elm;

  Prev(lcounter);
  Remove(llist, tmp);

  return elm;
}

/* Removes the current node of a counter. It shall only be used when
 * the counter is not at FIRST or LAST. The counter is set to the
 * following element.
 */
void *RemoveMoveRight(LList *llist, LListCounter *lcounter)
{
  LListNode *tmp = lcounter->current;
  void *elm = tmp->elm;

  Next(lcounter);
  Remove(llist, tmp);

  return elm;
}

int Length(LList *llist)
{
  return llist->count;
}
