/* loess.c: Implementation of functions to do one dimensional loess fitting
 */
/* Rune Lyngs, April 2003 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include "loess.h"
#include "iploess.h"
#include "polynomial.h"
#include "mergesort.h"
#include "rankselect.h"
#include "wrappers.h"

/* xmalloc(n): Wrapper function for malloc checking return value */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "loess");
#else
  return wmalloc(n);
#endif
}

/* cmp_doublep(x, y): compares the doubles pointed to by x and y */
static int cmp_doublep(double *x, double *y)
{
  return *x < *y;
}

/* cmp_doublepp(x, y): compares the doubles pointers pointed to by x and y */
static int cmp_doublepp(double **x, double **y)
{
  return **x < **y;
}

/* compute_q(n, alpha, lambda): returns the number of points used for
 * function inference with n datapoints and parameters alpha and
 * lambda.
 */
static int compute_q(int n, double alpha, int lambda)
{
  int q = floor(n * alpha);

  /* The number should be larger than the degree of the polynomial,
   * but no larger than the total number of points.
   */
  if (q < lambda + 1)
    q = lambda + 1;
  if (q > n)
    q = n;

  return q;
}

/* loess(n, x, y, alpha, lambda): set up structure for computing value
 * of the loess fitting based on the n points given by arrays x and y,
 * with parameters alpha and lambda.
 */
Loess *loess(int n, double *x, double *y, double alpha, int lambda)
{
  double **tmp = (double **)xmalloc(n * sizeof(double *));
  Loess *l = (Loess *)xmalloc(sizeof(Loess));
  int i;

  /* Create an array of pointers to the elements of x */
  for (i = 0; i < n; i++)
    tmp[i] = x + i;

  /* Sort this array */
  mergesort(tmp, n, sizeof(double *), (int (*)(void *, void *))cmp_doublepp);

  /*for (i = 0; i < n; i++)
    printf("%f ", *tmp[i]);
    exit(0);*/

  /* Initialise loess structure with data points sorted according to x
   * value.
   */
  l->x = (double *)xmalloc(n * sizeof(double));
  l->y = (double *)xmalloc(n * sizeof(double));
  for (i = 0; i < n; i++){
    l->x[i] = x[tmp[i] - x];
    l->y[i] = y[tmp[i] - x];
  }
  l->n = n;
  l->w = NULL;
  l->alpha = alpha;
  l->lambda = lambda;
  l->q = compute_q(n, alpha, lambda);

  /* Clean up */
  wfree(tmp);

  return l;
}

/* wloess(n, x, y, w, alpha, lambda): similar to loess, but returns
 * astructure also containing the weights.
 */
Loess *wloess(int n, double *x, double *y, double *w, double alpha, int lambda)
{
  double **tmp = (double **)xmalloc(n * sizeof(double *));
  Loess *l = (Loess *)xmalloc(sizeof(Loess));
  int i;

  /* Create an array of pointers to the elements of x */
  for (i = 0; i < n; i++)
    tmp[i] = x + i;

  /* Sort this array */
  mergesort(tmp, n, sizeof(double *), (int (*)(void *, void *))cmp_doublepp);

  /* Initialise loess structure with data points sorted according to x
   * value.
   */
  l->n = n;
  l->x = (double *)xmalloc(n * sizeof(double));
  l->y = (double *)xmalloc(n * sizeof(double));
  l->w = (double *)xmalloc(n * sizeof(double));
  for (i = 0; i < n; i++){
    l->x[i] = x[tmp[i] - x];
    l->y[i] = y[tmp[i] - x];
    l->w[i] = w[tmp[i] - x];
  }
  l->alpha = alpha;
  l->lambda = lambda;
  l->q = compute_q(n, alpha, lambda);

  /* Clean up */
  wfree(tmp);

  return l;
}

/* interval_start(x, q, n, d): finds an interval that contains q of
 * the n points in d closest to x.
 */
static int interval_start(double x, int q, int n, double *d)
{
  int i = 0, stepsize = n - q;

  while (stepsize > 0){
    if ((i + q < n) && (2 * x - d[i] > d[i + q])){
      /* Move interval to the right */
      i += stepsize;
      /* Make sure we stay within bounds */
      if (i + q > n)
	i = n - q;
    }
    else if ((i > 0) && (2 * x - d[i - 1] < d[i + q - 1])){
      /* Move interval to the left */
      i -= stepsize;
      /* Make sure we stay within bounds */
      if (i < 0)
	i = 0;
    }
    else
      /* Current interval should be OK */
      stepsize = 0;
    /* Halve stepsize, rounding up */
    stepsize = (stepsize + 1) / 2;
  }

  return i;
}

/* max(a, b): returns larger of a and b */
static double max(double a, double b)
{
  return (a > b ? a : b);
}

/* cube(x): computes the cube of x */
static double cube(double x)
{
  return x * x * x;
}

/* tricube(u, t): computes the tricube function */
static double tricube(double u, double t)
{
  if (u < t)
    return cube(1 - cube(u / t));
  else
    return 0;
}

/* loess_evaluate(x, l): evaluates the loess curve obtained from the
 * data in l at the point a.
 */
double loess_evaluate(Loess *l, double a)
{
  Polynomial *p;
  double deltaq, r, *w = (double *)xmalloc(l->q * sizeof(double));
  int i, b;

  if (l->alpha < 1){
    /* Determine the left boundary of interval of points that should be
     * used for inference.
     */
    b = interval_start(a, l->q, l->n, l->x);
    deltaq = max(a - l->x[b], l->x[b + l->q - 1] - a);
  }
  else{
    b = 0;
    deltaq = max(a - l->x[0], l->x[l->n - 1] - a);
  }

  /* Compute weights in w */
  /* First handle points with x value smaller than a */
  for (i = 0; i < l->q; i++){
    if (l->x[b + i] >= a) break;
    w[i] = tricube(a - l->x[b + i], deltaq);
    if (l->w != NULL)
      /* Weighted data */
      w[i] *= l->w[b + i];
  }
  /* Then handle points with x value larger than a */
  for (;i < l->q; i++){
    w[i] = tricube(l->x[b + i] - a, deltaq);
    if (l->w != NULL)
      /* Weighted data */
      w[i] *= l->w[b + i];
  }

  /* Find optimal polynomial and its value at a */
  p = polynomial_wlsdpolynomial(l->lambda + 1, l->q, l->x + b, l->y + b, w,
				0, l->x[b], l->x[b + l->q - 1]);
  if (p != NULL){
    r = polynomial_evaluate(p, a);
    polynomial_free(p);
  }
  else
    r = DBL_MAX;

  /* Clean up */
  wfree(w);

  return r;
}

/* square(a); returns a squared */
static double square(double a)
{
  return a * a;
}

/* bisquare(u, b): computes the bisquare function */
static double bisquare(double u, double b)
{
  if (u < b)
    return square(1 - square(u / b));
  else
    return 0;
}

/* loess_robustify(l): robustifies l by changing the weights */
Loess *loess_robustify(Loess *l)
{
  double s, m, *epsilon = (double *)xmalloc(l->n * sizeof(double));
  int i, j, k, c;
  Loess *lo = (Loess *)xmalloc(sizeof(Loess));

  /* Create a temporary Loess structure, initialise to copy of l */
  lo->x = (double *)xmalloc(l->n * sizeof(double));
  lo->y = (double *)xmalloc(l->n * sizeof(double));
  lo->w = (double *)xmalloc(l->n * sizeof(double));
  lo->alpha = l->alpha;
  lo->lambda = l->lambda;
  lo->q = l->q;
  lo->n = l->n;
  for (i = 0; i < l->n; i++){
    lo->x[i] = l->x[i];
    lo->y[i] = l->y[i];
    lo->w[i] = (l->w != NULL ? l->w[i] : 1);
  }

  /* Iteratively update the weights */
  j = 0;
  do{
    /* Compute deviation to current estimate for each point in l */
    k = 0;
    for (i = 0; i < l->n; i++){
      epsilon[i] = loess_evaluate(lo, l->x[i]);
      if (epsilon[i] != DBL_MAX)
	epsilon[i] = fabs(epsilon[i] - l->y[i]);
      else
	/* Could not evaluate the loess curve at i'th point */
	k++;
    }
    /* Compute robustness weights; we rescale the robustness weights
     * to have average 1, so we also need to compute the sum of the
     * robustness weights.
     */
    m = 6 * epsilon[rankselect(epsilon, (l->n - k) / 2, l->n, sizeof(double),
			       (int (*)(void *, void *))cmp_doublep)];
    s = 0;
    for (i = 0; i < l->n; i++){
      if (epsilon[i] != DBL_MAX){
	epsilon[i] = bisquare(epsilon[i], m);
	s += epsilon[i];
      }
      else
	epsilon[i] = 0;
    }

    if (s == 0){
      /* All weights are 0 - abort */
      loess_free(lo);
      wfree(epsilon);
      return NULL;
    }

    /* Update weights and check for convergence */
    c = 0;
    k = 0;
    for (i = 0; i < l->n; i++){
      /* Compute new weight for point i */
      m = epsilon[i] * l->n / s;
      if (l->w != NULL)
	m *= l->w[i];
      if (m == 0){
	/* Weight of this point is zero; do not include it in next iteration */
	k++;
      }
      else{
	/* Compare weight to old weight and update */
	if (fabs(m - lo->w[i - k]) > _LOESS_ROBUSTSTOP)
	  /* This weight has not converged yet */
	  c = 1;
	lo->w[i - k] = m;
	/* Convergence not attained if there is a change in the set of
	 * points with non-zero weight.
	 */ 
	if (lo->x[i - k] != l->x[i]){
	  lo->x[i - k] = l->x[i];
	  c = 1;
	}
	lo->y[i - k] = l->y[i];
      }
    }
    /* Convergence not attained if there is a change in the set of
     * points with non-zero weight.
     */
    if (lo->n != l->n - k){
      lo->n = l->n - k;
      c = 1;
    }
    lo->q = compute_q(lo->n, lo->alpha, lo->lambda);
    /* Another iteration older... */
    j++;
  } while (c && (j < _LOESS_ROBUSTMAXITERATIONS));

  /* Clean up */
  wfree(epsilon);
  wfree(l->x);
  wfree(l->y);
  if (l->w != NULL)
    wfree(l->w);

  /* Copy data from lo to l */
  l->x = lo->x;
  l->y = lo->y;
  l->w = lo->w;
  l->n = lo->n;
  l->q = lo->q;
  wfree(lo);

  return l;
}

/* loess_nisd(l1, l2, a, b, n): approximates the integral of square
 * differences between loess fits l1 and l2 in the interval from a to
 * b by dividing the range into n subintervals, approximating each
 * loess fit with a polynomial of the same order in each subinterval,
 * and summing the integrals of squared polynomial differences for
 * each subinterval.
 */
double loess_nisd(Loess *l1, Loess *l2, double a, double b, int n)
{
  double result;
  IPLoess *ipl1, *ipl2;

  /* Convert loess fits to interpolated loess fits */
  ipl1 = iploess(l1, n, a, b);
  ipl2 = iploess(l2, n, a, b);

  /* Compute integral for interpolated loess fits */
  result = iploess_isd(ipl1, ipl2, a, b);

  /* Clean up */
  iploess_free(ipl1);
  iploess_free(ipl2);

  return result;
}

/* loess_changealpha(alpha, l): changes the alpha value of the loess
 * structure l to alpha.
 */
void loess_changealpha(double alpha, Loess *l)
{
  l->alpha = alpha;
  l->q = compute_q(l->n, alpha, l->lambda);
}

/* loess_changelambda(lambda, l): changes the lambda value of the
 * loess structure l to lambda.
 */
void loess_changelambda(int lambda, Loess *l)
{
  l->lambda = lambda;
  l->q = compute_q(l->n, l->alpha, lambda);
}

/* loess_free(l): frees memory used by the loess structure referenced by l */
void loess_free(Loess *l)
{
  wfree(l->x);
  wfree(l->y);
  if (l->w != NULL)
    wfree(l->w);
  wfree(l);
}
