/* select.c: Implementation of an efficient (linear time) method for
 * finding the k'th smallest element in an array.
 */
/* Rune Lyngs, March 2003 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rankselect.h"
#include "wrappers.h"

/* xmalloc(n): Tries to allocate n bytes of memory; if unsuccessful an
 * error is reported.
 */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "rankselect");
#else
  return wmalloc(n);
#endif
}

/* fivemedian(data, less_than): Find the median of the five items
 * referenced by the pointers in data according to the ordering
 * imposed by less_than, and rearrange the pointers such that the
 * median is in position 2, the two items smaller than the median in
 * positions 0 and 1, and the two items larger than median in
 * positions 3 and 4. The function uses exactly 6 item comparisons,
 * and at most 6 assignments to swap the pointers.
 */
static void fivemedian(void **data, int (*less_than)(void *, void *))
{
  void *tmp;
  int l[6];
  int i;

  /* Comparison 1 */
  /* Compare first and second element */
  if (less_than(data[1], data[0]))
    /* data[0] > data[1] */
    l[0] = 0;
  else
    /* data[1] >= data[0] */
    l[0] = 1;
 
  /* Comparison 2 */
  /* Compare third and fourth element */
  if (less_than(data[3], data[2]))
    /* data[2] > data[3] */
    l[1] = 2;
  else
    /* data[3] >= data[2] */
    l[1] = 3;

  /* Comparison 3 */
  /* Compare the two items that were larger in previous comparisons */
  if (less_than(data[l[1]], data[l[0]])){
    /* Item of first comparison larger */
    l[2] = l[0];
    /* We no longer need l[0] or l[1], but we will need the looser of
     * this comparison; record in l[0].
     */
    l[0] = l[1];
  }
  else
    /* Item of second comparison larger */
    l[2] = l[1];

  /* Comparison 4 */
  /* Compare smaller item of first comparison with largest item so far
   * to fifth item.
   */
  if (less_than(data[4], data[l[2] ^ 1]))
    /* Fifth item is smaller */
    l[3] = l[2] ^ 1;
  else
    /* Fifth item is larger */
    l[3] = 4;

  /* Comparison 5 */
  /* Compare larger item in previous comparison to smaller of the two
   * items in comparison three.
   */
  if (less_than(data[l[3]], data[l[0]])){
    /* Item from third comparison is larger; this is too large to be
     * median and the smaller item in comparison four is too small to
     * be median.
     */
    l[4] = l[0];
    /* Use l[1] to record item known to be smaller than median */
    l[1] = l[2] ^ l[3] ^ 5;
    /* Use l[5] to record second item (looser of looser in first three
     * comparisons) of final comparison.
     */
    l[5] = l[0] ^ 1;
  }
  else{
    /* Item from previous comparison is larger; this is too large to
     * be median and the item that in comparison one or two was
     * determined to be smaller than the smaller item of comparison
     * three is too small to be median.
     */
    l[4] = l[3];
    /* Use l[1] to record item known to be smaller than median */
    l[1] = l[0] ^ 1;
    /* Use l[5] to record second item (smaller item in comparison
     * four) of final comparison.
     */
    l[5] = l[2] ^ l[3] ^ 5;
  }

  /* Comparison 6 */
  /* Compare two remaining items - larger one is median */
  if (less_than(data[l[5]], data[l[0] ^ l[3] ^ l[4]]))
    /* Smaller item from fifth comparison is median */
    l[5] = l[0] ^ l[3] ^ l[4];
  /* Otherwise, the item already recorded in l[5] is median */

  /* l[2] and l[4] holds the two items larger than the median, l[1]
   * holds one item smaller than the median, and l[5] holds the
   * median; rearrange the array to have the median at position 0, the
   * two items smaller than the median in positions 1 and 2, and the
   * two items larger than the median in positions 3 and 4.
   */
  /* Start by making sure that l[2] < l[4], and for convenience move
   * them to l[3] and l[4].
   */
  if (l[2] < l[4])
    l[3] = l[2];
  else{
    l[3] = l[4];
    l[4] = l[2];
  }
  /* Then store the positions of the two elements smaller than the
   * median in l[1] and l[2], with the leftmost position in l[1].
   */
  if (10 - l[3] - l[4] - l[5] > 2 * l[1])
    l[2] = 10 - l[3] - l[4] - l[5] - l[1];
  else{
    l[2] = l[1];
    l[1] = 10 - l[3] - l[4] - l[5] - l[1];
  }

  /* Detect items that are already in a proper place */
  for (i = 1; i < 3; i++)
    if (l[i] < 2)
      /* Element referred to by l[i] is already in place */
      l[i] = -1;
  for (; i < 5; i++)
    if (l[i] > 3)
      /* Element referred to by l[i] is already in place */
      l[i] = -1;

  if (l[5] != 2){
    /* Median is not in place - move it to tmp */
    tmp = data[l[5]];
    i = l[5];
    l[0] = -1;
  }
  else if (l[3] != -1){
    /* leftmost of items larger than median not in place - move it to tmp */
    tmp = data[l[3]];
    i = l[3];
    l[3] = -1;
  }
  else
    /* Everything is already in place */
    return;

  for (;;){
    if (i < 3){
      /* Place just vacated should hold an item smaller than the
       * median, or the median (that was stored in tmp).
       */
      if (l[2] >= 0){
	data[i] = data[l[2]];
	i = l[2];
	l[2] = -1;
      }
      else if (l[1] >= 0){
	data[i] = data[l[1]];
	i = l[1];
	l[1] = -1;
      }
      else
	break;
    }
    else{
      /* Place just vacated should hold an item larger than the median */
      if (l[3] >= 0){
	data[i] = data[l[3]];
	i = l[3];
	l[3] = -1;
      }
      else if (l[4] >= 0){
	data[i] = data[l[4]];
	i = l[4];
	l[4] = -1;
      }
      else
	break;
    }
  }
  /* We still need to move the item from the first position vacated to
   * the last position vacated.
   */
  data[i]= tmp;
}

/* rankselect_min(data, n, less_than): returns a minimum element,
 * according to less_than, of the n items pointed to by data.
 */
static void *rankselect_min(void **data, int n,
			    int (*less_than)(void *, void *))
{
  int i;
  void *m = data[0];

  /* Starting with the first item being the current minimum, run
   * through the remaining items updating m whenever an item is
   * smaller than the current minimum.
   */
  for (i = 1; i < n; i++)
    if (less_than(data[i], m))
      m = data[i];

  return m;
}

/* rankselect_min2(data, n, size, less_than): returns a second smallest
 * element, according to less_than, of the n items pointed to by data.
 */
static void *rankselect_min2(void **data, int n,
			     int (*less_than)(void *, void *))
{
  int i;
  void *m1, *m2;

  /* Find minimum and second smallest item of first two items */
  if (less_than(data[1], data[0])){
    m1 = data[1];
    m2 = data[0];
  }
  else{
    m1 = data[0];
    m2 = data[1];
  }
  /* Run through the remaining items, updating m1 and m2 */
  for (i = 2; i < n; i++)
    if (less_than(data[i], m2)){
      if (less_than(data[i], m1)){
	m2 = m1;
	m1 = data[i];
      }
      else
	m2 = data[i];
    }

  return m2;
}

/* rankselect_max(data, n, size, less_than): returns a maximum
 * element, according to less_than, of the n items pointed to by data.
 */
static void *rankselect_max(void **data, int n,
			    int (*less_than)(void *, void *))
{
  int i;
  void *m = data[0];

  /* Starting with the first item being the current maximum, run
   * through the remaining items updating m whenever current maximum
   * is surpassed.
   */
  for (i = 1; i < n; i++)
    if (less_than(m, data[i]))
      m = data[i];

  return m;
}

/* rankselect_max2(data, n, size, less_than): returns a second largest
 * element, according to less_than, of the n items pointed to by data.
 */
static void *rankselect_max2(void **data, int n,
			     int (*less_than)(void *, void *))
{
  int i;
  void *m1, *m2;

  /* Find maximum and second largest item of first two items */
  if (less_than(data[1], data[0])){
    m1 = data[0];
    m2 = data[1];
  }
  else{
    m1 = data[1];
    m2 = data[0];
  }
  /* Run through the remaining items, updating m1 and m2 */
  for (i = 2; i < n; i++)
    if (less_than(m2, data[i])){
      if (less_than(m1, data[i])){
	m2 = m1;
	m1 = data[i];
      }
      else
	m2 = data[i];
    }

  return m2;
}

/* _rankselect(data, k, n, less_than): returns an element of rank k
 * among the n items pointed to by data according to the ordering
 * imposed by less_than; the pointers in data might be rearranged, but
 * the addresses pointed to are not modified.
 */
static void *_rankselect(void **data, int k, int n,
			  int (*less_than)(void *, void *))
{
  int i, j, s, *c;
  void **tmp;
  void *m;


  /* Handle a few special cases */
  switch(k){
  case 0:
    return rankselect_min(data, n, less_than);
  case 1:
    return rankselect_min2(data, n, less_than);
  }
  switch(n - k){
  case 1:
    return rankselect_max(data, n, less_than);
  case 2:
    return rankselect_max2(data, n, less_than);
  }
  /* We now know that we are not looking for smallest, second
   * smallest, largest, or second largest item; hence there must be at
   * least 5 items.
   */
  tmp = (void **)xmalloc(n / 5 * sizeof(void *));

  /* Find median of medians of blocks of 5 items */
  for (i = 0; i < n / 5; i++){
    fivemedian(data + i * 5, less_than);
    tmp[i] = data[2 + i * 5];
  }
  m = _rankselect(tmp, n / 10, n / 5, less_than);
  wfree(tmp);

  /* Split data according to m, and count number of elements smaller
   * than, equal to, and larger than m.
   */
  s = 0;
  c = (int *)xmalloc(n * sizeof(int));
  for (i = 0; i <= n - 5; i += 5)
    /* Start by comparing the median of each block of 5 items to m;
     * these comparison will give information about two other items in
     * the block.
     */
    if (less_than(data[2 + i], m)){
      /* Median of block and the two elements smaller than the median
       * are smaller than m.
       */
      s += 3;
      for (j = 0; j < 3; j++)
	c[i + j] = 0;
      /* Compare last two elements of block to m */
      for (; j < 5; j++)
	if (less_than(m, data[i + j]))
	  /* This item is larger than m */
	  c[i + j] = 1;
	else{
	  /* Item is no larger than m */
	  s++;
	  c[i + j] = 0;
	}
    }
    else{
      /* Median of block and the two elements larger than the median
       * are at least as large as m.
       */
      for (j = 4; j > 1; j--)
	c[i + j] = 1;
      /* Compare last two elements of block to m */
      for (; j >= 0; j--)
	if (less_than(m, data[i + j]))
	  /* This item is larger than m */
	  c[i + j] = 1;
	else{
	  /* Item is no larger than m */
	  s++;
	  c[i + j] = 0;
	}
    }

  /* There might be up to 4 elements that have not been handled as
   * part of a block of 5 items.
   */
  for (; i < n; i++)
    if (less_than(m, data[i]))
      /* This item is larger than m */
      c[i] = 1;
    else{
      /* This item is no larger than m */
      s++;
      c[i] = 0;
    }

  if (k < s){
    /* The item of rank k is among the items smaller than m */
    /* Copy all items smaller than m to tmp */
    tmp = (void **)xmalloc(s * sizeof(void *));
    j = 0;
    for (i = 0; i < n; i++)
      if (!c[i]){
	tmp[j] = data[i];
	j++;
      }
    /* Find the item of rank k among these */
    m = _rankselect(tmp, k, s, less_than);
  }
  else{
    /* The item of rank k is among the items larger than m */
    /* Copy all items larger than m to tmp */
    tmp = (void **)xmalloc((n - s) * sizeof(void *));
    j = 0;
    for (i = 0; i < n; i++)
      if (c[i]){
	tmp[j] = data[i];
	j++;
      }
    /* Find the item of rank k - s among these */
    m = _rankselect(tmp, k - s, n - s, less_than);
  }

  /* Clean up */
  wfree(tmp);
  wfree(c);

  return m;
}

/* rankselect(data, k, n, size, less_than): returns the index of an
 * element of rank k among the n items in data according to the
 * ordering imposed by less_than; each item is assumed to require size
 * bytes of memory, but a work copy of data is made such that data is
 * not modified by a call to this function.
 */
int rankselect(void *data, int k, int n, size_t size,
	       int (*less_than)(void *, void *))
{
  int i;
  void *m, **tmp;

  /* Check consistency */
  if ((n <= 0) || (k < 0) || (k >= n)){
    fprintf(stderr, "Cannot select item of rank %d among %d elements.\n",
	    k, n);
    return -1;
  }

  /* Make a work copy of pointers to the data */
  tmp = (void **)xmalloc(n * sizeof(void *));
  for (i = 0; i < n; i++)
    tmp[i] = data + i * size;

  /* Find item of rank k in data */
  m = _rankselect(tmp, k, n, less_than);
  /* Clean up */
  wfree(tmp);

  return (m - data) / size;
}
