/* relative_rank_selection.c: Implementations of functions for
 * computing normalising functions based on relative rank selection
 * invariant sets. The two main functions compute normalising
 * functions for all possible thresholds and for a specific threshold,
 * respectively.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "relative_rank_selection.h"
#include "mergesort.h"
#include "wrappers.h"

/* xmalloc(n): wrapper function for malloc, testing whether memory was
 * correctly allocated.
 */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "relative_rank_selection");
#else
  return wmalloc(n);
#endif
}

/* cmp_idpair(a, b): return true if value of a is less than value of b */
static int cmp_idpair(_RRS_idpair *a, _RRS_idpair *b)
{
  return (a->value < b->value);
}

/* cmp_iipair(a, b): return true if value of a is less than value of b */
static int cmp_iipair(_RRS_iipair *a, _RRS_iipair *b)
{
  return (a->value > b->value);
}

/* maximum_set_size(n, x, y): for each of the n points given by x and
 * y computes the size of the maximum set with conserved relative
 * ordering that contains the element.
 */
static int *maximum_set_size(int n, double *x, double *y)
{
  _RRS_idpair *sort = (_RRS_idpair *)xmalloc(n * sizeof(_RRS_idpair));
  double *min = (double *)xmalloc(n * sizeof(double));
  int *ss = (int *)xmalloc(n * sizeof(int));
  int i, max, u, l;

  /* Sort elements according to x value */
  for (i = 0; i < n; i++){
    sort[i].index = i;
    sort[i].value = x[i];
  }
  mergesort((void *)sort, n, sizeof(_RRS_idpair),
	    (int (*)(void *, void *))cmp_idpair);

  /* For each element, find size of maximum ordered set having that
   * element as largest member.
   */
  max = 0;
  for (i = 0; i < n; i++){
    l = 0; u = max;
    /* Search binarily for maximum size set we can add element i to */
    while (u > l){
      if (y[sort[i].index] >= min[(u + l) / 2])
	/* We can extend set of size (u + l) / 2 + 1 with element i */
	l = (u + l) / 2 + 1;
      else
	/* We cannot extend set of size (u + l) / 2 + 1 with element i */
	u = (u + l) / 2;
    }
    ss[sort[i].index] = l + 1;
    /* Update array of minimum maximum members sets of a particular
     * size can have.
     */
    if (l == max)
      max++;
    min[l] = y[sort[i].index];
  }

  /* For each element, find size of maximum ordered set having that
   * element as smallest member.
   */
  max = 0;
  for (i = n - 1; i >= 0; i--){
    l = 0; u = max;
    /* Search binarily for maximum size set we can add element i to */
    while (u > l){
      if (y[sort[i].index] <= min[(u + l) / 2])
	/* We can extend set of size (u + l) / 2 + 1 with element i */
	l = (u + l) / 2 + 1;
      else
	/* We cannot extend set of size (u + l) / 2 + 1 with element i */
	u = (u + l) / 2;
    }
    ss[sort[i].index] += l;
    /* Update array of minimum maximum members sets of a particular
     * size can have.
     */
    if (l == max)
      max++;
    min[l] = y[sort[i].index];
  }

  /* Clean up */
  wfree(min);
  wfree(sort);

  return ss;
}

/* relative_rank_selection_all(n, x, y, new_function): finds
 * normalising function returned by new_function when only points that
 * are member of an ordered set at most a given threshold smaller than
 * a maximum ordered set are used for all possible thresholds. Only
 * for thresholds where more points are included will the
 * corresponding entry in the array of normalising functions returned
 * contain an actual function, all other entries in the array are
 * NULL.
 */
void **relative_rank_selection_all(int n, double *x, double *y,
				   _RRS_newfunction new_function)
{
  _RRS_iipair *sort = (_RRS_iipair *)xmalloc(n * sizeof(_RRS_iipair));
  int *ss = maximum_set_size(n, x, y);
  double *xs, *ys;
  int i;
  void **fs = (void **)xmalloc(n * sizeof(void *));

  /* Sort the points according to maximum set sizes, and initialise
   * array of normalising functions.
   */
  for (i = 0; i < n; i++){
    sort[i].index = i;
    sort[i].value = ss[i];
    fs[i] = NULL;
  }
  wfree(ss);
  mergesort((void *)sort, n, sizeof(_RRS_iipair),
	    (int (*)(void *, void *))cmp_iipair);
  if (new_function != NULL){
    /* Store the points in xs, ys in order of decreasing set sizes, and
     * compute the next normalising function whenever the maximum set
     * size decreases. The normalising function at position i in the
     * array of normalising functions is the function returned by
     * new_function when including all points that are members of an
     * ordered set at most i elements smaller than a maximum ordered
     * set.
     */
    xs = (double *)xmalloc(n * sizeof(double));
    ys = (double *)xmalloc(n * sizeof(double));
    /* There are at least as many elements that are members of a maximum
     * ordered set as the size of a maximum ordered set.
     */
    for (i = 0; i < sort[0].value; i++){
      xs[i] = x[sort[i].index];
      ys[i] = y[sort[i].index];
    }
    for (; i < n; i++){
      if (sort[i].value < sort[i - 1].value)
	/* Change in set size - compute normalising function for all
	 * preceeding points.
	 */
	fs[sort[0].value - sort[i - 1].value] = new_function(i, xs, ys);

      xs[i] = x[sort[i].index];
      ys[i] = y[sort[i].index];
    }

    /* Compute normalising function when all points are included */
    fs[sort[0].value - sort[n - 1].value] = new_function(n, xs, ys);

    /* Clean up */
    wfree(xs);
    wfree(ys);
  }
  else{
    /* Store arrays in fs where the t'th array has i'th element equal
     * to 1 if the i'th point is a member of an ordered set at most t
     * elements smaller than a maximum ordered set, 0 otherwise; if
     * the t'th and t - 1'st array would be identical, the t'th
     * element of fs is NULL.
     */
    /* Start with initialising all entries to 0 */
    xs = (double *)xmalloc(n * sizeof(double));
    for (i = 0; i < n; i++)
      xs[i] = 0;
    /* Change values to 1 in order of increasing rank difference;
     * whenever the rank difference increases, store the array at the
     * right position in fs and allocate a new array with the same
     * content.
     */
    /* There are at least as many elements that are members of a maximum
     * ordered set as the size of a maximum ordered set.
     */
    for (i = 0; i < sort[0].value; i++)
      xs[sort[i].index] = 1;
    for (; i < n; i++){
      if (sort[i].value < sort[i - 1].value){
	/* Change in set size - store array in fs */
	ys = (double *)xmalloc(n * sizeof(double));
	memcpy(ys, xs, n * sizeof(double));
	fs[sort[0].value - sort[i - 1].value] = (void *)ys;
      }
      xs[sort[i].index] = 1;
    }
    /* We haven't stored last invariant set (i.e. the set consisting
     * of all points) in fs yes.
     */
    fs[sort[i].value] = (void *)xs;
  }

  /* Clean up */
  wfree(sort);

  return fs;
}

/* relative_rank_selection(n, x, y, t, new_function): computes
 * normalising function by passing the points that are members of an
 * ordered set at most t elements smaller than a maximum ordered set
 * to new_function.
 */
void *relative_rank_selection(int n, double *x, double *y, int t,
			      _RRS_newfunction new_function)
{
  void *f;
  int *ss = maximum_set_size(n, x, y);
  double *xs, *ys;
  int i, j, m, mss;

  /* Find size of maximum ordered set */
  mss = ss[0];
  for (i = 1; i < n; i++)
    if (ss[i] > mss) mss = ss[i];

  if (new_function != NULL){
    /* Find number of points that are member of an ordered set at most t
     * elements smaller than a maximum ordered set.
     */
    m = (ss[0] >= mss - t);
    for (i = 1; i < n; i++)
      if (ss[i] >= mss - t) m++;

    /* Copy points that are member of an ordered set at most t elements
     * smaller than a maximum ordered set to xs, ys.
     */
    xs = (double *)xmalloc(m * sizeof(double));
    ys = (double *)xmalloc(m * sizeof(double));
    i = 0;
    for (j = 0; j < m; j++){
      /* Skip points that are not member of an ordered set at most t
       * elements smaller than a maximum ordered set.
       */
      while (ss[i] < mss - t) i++;
      xs[j] = x[i];
      ys[j] = y[i];
      i++;
    }

    /* Compute normalising function based on points xs, ys */
    f = new_function(m, xs, ys);

    /* Clean up */
    wfree(ss);
    wfree(xs);
    wfree(ys);

    return f;
  }
  else{
    /* Create array with i'th element 1 if the i'th point is member of
     * an ordered set at most t elements smaller than a maximum
     * ordered set, 0 otherwise.
     */
    xs = (double *)xmalloc(n * sizeof(double));
    for (i = 0; i < n; i++)
      xs[i] = 0;
    for (i = 0; i < n; i++)
      if (ss[i] >= mss - t) xs[i] = 1;

    /* Clean up */
    wfree(ss);

    return (void *)xs;
  }
}
