/* reliability.c: implementation of function for computing a
 * normalising function based on reliability. The reliability of a
 * point is determined as the probability of observing the point when
 * any subset of n points with conserved relative ordering is assigned
 * an energy proportional to -n and all subsets where relative
 * ordering is not conserved are assigned infinite energy.
 */
/* Rune Lyngs, February 2003 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "reliability.h"
#include "mergesort.h"
#include "splay.h"
#include "gmp_interface.h"
#include "wrappers.h"

/* xmalloc(n): wrapper function for malloc, testing whether memory was
 * correctly allocated.
 */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "reliability");
#else
  return wmalloc(n);
#endif
}

/* cmp_idpair(a, b): return true if value of a is less than value of b */
static int cmp_idpair(_RELIABILITY_idpair *a, _RELIABILITY_idpair *b)
{
  return (a->value < b->value);
}

/* cmp_treeitem(a, b): return true if a has smaller key than b */
static int cmp_treeitem(_RELIABILITY_treeitem *a, _RELIABILITY_treeitem *b)
{
  if (a->y == b->y){
    if (a->x == b->x)
      return (a->i < b->i);
    else
      return (a->x < b->x);
  }
  else
    return (a->y < b->y);
}

/* update_treeitem(a, b, c): updates aggregate value of item a when
 * items at roots of left and right subtrees are b and c.
 */
static void update_treeitem(_RELIABILITY_treeitem *a, _RELIABILITY_treeitem *b,
			    _RELIABILITY_treeitem *c)
{
  if (b != NULL){
    gmpi_add(a->agr_value, a->own_value, b->agr_value);
    if (c != NULL)
      gmpi_add(a->agr_value, a->agr_value, c->agr_value);
  }
  else if (c != NULL)
    gmpi_add(a->agr_value, a->own_value, c->agr_value);
  else
    gmpi_set(a->agr_value, a->own_value);
}

/* find_saggregate(i, l, r): determines the aggregate value of items no
 * larger than i, where l is the item of the left subtree of i and r
 * is the item of the right subtree of i when i has been rotated up as
 * root.
 */
/* To avoid having to initialise and clearing a multiprecission type
 * for returning the aggregate value, a global variable is declared to
 * hold the return value. Remember to initialise it before use and
 * clear it after use.
 */
static mp_type aggregate;
static void *find_saggregate(_RELIABILITY_treeitem *l,
			     _RELIABILITY_treeitem *i,
			     _RELIABILITY_treeitem *r)
{
  if (i == NULL){
    if (l == NULL)
      gmpi_set_ui(aggregate, 0);
    else
      gmpi_set(aggregate, l->agr_value);
  }
  else if (l == NULL)
    gmpi_set(aggregate, i->own_value);
  else
    gmpi_add(aggregate, i->own_value, l->agr_value);

  return &aggregate;
}

/* find_laggregate(i, l, r): determines the aggregate value of items no
 * smaller than i, where l is the item of the left subtree of i and r
 * is the item of the right subtree of i when i has been rotated up as
 * root.
 */
/* Reuse aggregate variable from find_saggregate for returnin value
 */
static void *find_laggregate(_RELIABILITY_treeitem *l,
			     _RELIABILITY_treeitem *i,
			     _RELIABILITY_treeitem *r)
{
  if (i == NULL){
    if (r == NULL)
      gmpi_set_ui(aggregate, 0);
    else
      gmpi_set(aggregate, r->agr_value);
  }
  else if (r == NULL)
    gmpi_set(aggregate, i->own_value);
  else
    gmpi_add(aggregate, i->own_value, r->agr_value);

  return &aggregate;
}

/* sum(a): adds a's own value to the global variable total */
static mp_type total;
static void sum(_RELIABILITY_treeitem *a)
{
  gmpi_add(total, total, a->own_value);
}

/* reliability(n, x, y, energy, new_function, transform): computes a
 * normalising function based on the weighted set of n points given by
 * x and y with weights obtained by passing reliabilities, with a per
 * member contribution of probability to the total probability of a
 * set with conserved relative ordering, to transform. The weighted
 * set is passed to new_function to obtain the normalising function.
 */
void *reliability(int n, double *x, double *y, double p,
		  _RELIABILITY_newfunction new_function,
		  _RELIABILITY_transform transform)
{
  void *f;
  int i;
  _RELIABILITY_idpair *sort
    = (_RELIABILITY_idpair *)xmalloc(n * sizeof(_RELIABILITY_idpair));
  splaytree *t
    = splay_create((int (*)(void *, void *))cmp_treeitem,
		   (void (*)(void *, void *, void *))update_treeitem);
  _RELIABILITY_treeitem *br
    = (_RELIABILITY_treeitem *)xmalloc(n * sizeof(_RELIABILITY_treeitem));
  _RELIABILITY_treeitem *item;
  mp_type *fr = (mp_type *)xmalloc(n * sizeof(mp_type));
  mp_type prob;
  double *w = (double *)xmalloc(n * sizeof(double));
  gmpi_init(aggregate);
  gmpi_init(prob);
#ifdef _RELIABILITY_NOROUNDING
  gmpi_set_d(prob, p);
#else
  /* We restrict precission of p to 1/100 to avoid time consuming
   * computations with rationals with large enumerators and
   * denominators.
   */
  gmpi_set_si(prob, floor(p * 100), 100);
#endif

  /* Sort elements according to x value (and initialise all mp_types
   * allocated).
   */
  for (i = 0; i < n; i++){
    sort[i].index = i;
    sort[i].value = x[i];
    gmpi_init(br[i].own_value);
    gmpi_init(br[i].agr_value);
    gmpi_init(fr[i]);
  }
  mergesort((void *)sort, n, sizeof(_RELIABILITY_idpair),
	    (int (*)(void *, void *))cmp_idpair);

  /* Process elements in order of increasing x value to compute
   * `forward' reliability.
   */
  for (i = 0; i < n; i++){
    br[i].y = y[sort[i].index];
    br[i].x = sort[i].value;
    br[i].i = sort[i].index;
    item = splay_accesssmaller(t, &br[i]);
    if (item == NULL)
      gmpi_set_d(br[i].own_value, p);
    else{
      splay_evaluate(t, item,
		     (void *(*)(void *, void *, void *))find_saggregate);
      gmpi_inc(aggregate);
      gmpi_set(br[i].own_value, aggregate);
      gmpi_mul(br[i].own_value, br[i].own_value, prob);
    }
    splay_insert(t, &br[i]);
  }
  for (i = 0; i < n; i++)
    gmpi_set(fr[i], br[i].own_value);

  /* Compute partition function */
  gmpi_init_set_ui(total, 0);
  splay_destroytree(t, (void (*)(void *))sum);

  /* Process elements in order of decreasing x value to compute
   * `backward' reliability.
   */
  t = splay_create((int (*)(void *, void *))cmp_treeitem,
		   (void (*)(void *, void *, void *))update_treeitem);
  for (i = n - 1; i >= 0; i--){
    item = splay_accesslarger(t, &br[i]);
    if (item == NULL)
      gmpi_set_ui(br[i].own_value, 1);
    else{
      splay_evaluate(t, item,
		     (void *(*)(void *, void *, void *))find_laggregate);
      gmpi_mul(aggregate, aggregate, prob);
      gmpi_inc(aggregate);
      gmpi_set(br[i].own_value, aggregate);
    }
    splay_insert(t, &br[i]);
  }
  splay_deletetree(t, NULL);

  /* Compute basic reliabilities */
  for (i = 0; i < n; i++){
    gmpi_mul(fr[i], fr[i], br[i].own_value);
    gmpi_div(fr[i], fr[i], total);
  }

  /* Transform reliabilities and convert to doubles */
  if (transform != NULL)
    transform(n, fr);
  for (i = 0; i < n; i++)
    w[sort[i].index] = gmpi_get_d(fr[i]);

  if (new_function != NULL){
    /* Compute normalising function */
    f = new_function(n, x, y, w);
    wfree(w);
  }
  else{
    /* Store weights in f for return */
    f = (void *)w;
  }

  /* Clean up */
  gmpi_clear(aggregate);
  gmpi_clear(prob);
  gmpi_clear(total);
  for (i = 0; i < n; i++){
    gmpi_clear(fr[i]);
    gmpi_clear(br[i].own_value);
    gmpi_clear(br[i].agr_value);
  }
  wfree(sort);
  wfree(br);
  wfree(fr);

  return f;
}
